import os
import pytest
import tempfile
import unittest.mock


from pipemake.pipemake import main


@pytest.mark.parametrize(
    "wildcard_str",
    [
        "tests/files/wildcardIO/{samples}_{reads}.fq.gz",
    ],
)
def test_pipemake_main_wildcard_wo_error(wildcard_str):
    test_dir = tempfile.mkdtemp()
    os.environ["PM_SNAKEMAKE_DIR"] = "tests/files/snakemakeIO"

    # Assign the workflor prefix
    workflow_prefix = os.path.join(test_dir, "test_workflow")

    # Assign the command line arguments
    test_cmd = [
        "",
        "fastq-filter",
        "--fastq-wildcard",
        wildcard_str,
        "--workflow-prefix",
        workflow_prefix,
        "--resource-yml",
    ]

    # Run pipemake with the test command
    with unittest.mock.patch("sys.argv", test_cmd):
        main()

    # Check if the workflow files were created
    assert os.path.isfile(f"{workflow_prefix}.smk")
    assert os.path.isfile(f"{workflow_prefix}.yml")
    assert os.path.isdir(f"{workflow_prefix}")
    assert os.path.isfile(f"{workflow_prefix}/pipemake/pipeline.log")
    assert os.path.isdir(f"{workflow_prefix}/pipemake")
    assert os.path.isfile(f"{workflow_prefix}/pipemake/backups/test_workflow.smk.bkp")
    assert os.path.isfile(f"{workflow_prefix}/pipemake/backups/test_workflow.yml.bkp")
    assert os.path.isfile(
        f"{workflow_prefix}/pipemake/backups/test_workflow.resources.yml.bkp"
    )
    assert os.path.isdir(f"{workflow_prefix}/pipemake/backups")
    assert os.path.isfile(f"{workflow_prefix}/pipemake/modules/fastq_filter_fastp.smk")
    assert not os.path.isfile(f"{workflow_prefix}/pipemake/modules/test_script.smk")
    assert os.path.isdir(f"{workflow_prefix}/pipemake/modules")
    assert os.path.isdir(f"{workflow_prefix}/FASTQ/Unfiltered")


@pytest.mark.parametrize(
    "table_str",
    [
        "tests/files/fileIO/test_table.tsv",
    ],
)
def test_pipemake_main_table_wo_error(table_str):
    test_dir = tempfile.mkdtemp()
    os.environ["PM_SNAKEMAKE_DIR"] = "tests/files/snakemakeIO"

    # Assign the workflor prefix
    workflow_prefix = os.path.join(test_dir, "test_workflow")

    # Assign the command line arguments
    test_cmd = [
        "",
        "fastq-filter",
        "--fastq-table",
        table_str,
        "--workflow-prefix",
        workflow_prefix,
        "--resource-yml",
    ]

    # Run pipemake with the test command
    with unittest.mock.patch("sys.argv", test_cmd):
        main()

    # Check if the workflow files were created
    assert os.path.isfile(f"{workflow_prefix}.smk")
    assert os.path.isfile(f"{workflow_prefix}.yml")
    assert os.path.isdir(f"{workflow_prefix}")
    assert os.path.isfile(f"{workflow_prefix}/pipemake/pipeline.log")
    assert os.path.isdir(f"{workflow_prefix}/pipemake")
    assert os.path.isfile(f"{workflow_prefix}/pipemake/backups/test_workflow.smk.bkp")
    assert os.path.isfile(f"{workflow_prefix}/pipemake/backups/test_workflow.yml.bkp")
    assert os.path.isfile(
        f"{workflow_prefix}/pipemake/backups/test_workflow.resources.yml.bkp"
    )
    assert os.path.isdir(f"{workflow_prefix}/pipemake/backups")
    assert os.path.isfile(f"{workflow_prefix}/pipemake/modules/fastq_filter_fastp.smk")
    assert os.path.isfile(f"{workflow_prefix}/pipemake/modules/test_script.smk")
    assert os.path.isdir(f"{workflow_prefix}/pipemake/modules")
    assert os.path.isdir(f"{workflow_prefix}/FASTQ/Unfiltered")

    with open(f"{workflow_prefix}.yml") as f:
        for line in f:
            print(line)


@pytest.mark.parametrize(
    "table_str",
    [
        "tests/files/fileIO/test_table2.tsv",
    ],
)
def test_pipemake_main_table_w_error(table_str):
    test_dir = tempfile.mkdtemp()
    os.environ["PM_SNAKEMAKE_DIR"] = "tests/files/snakemakeIO"

    # Assign the workflor prefix
    workflow_prefix = os.path.join(test_dir, "test_workflow")

    # Assign the command line arguments
    test_cmd = [
        "",
        "fastq-filter",
        "--fastq-table",
        table_str,
        "--workflow-prefix",
        workflow_prefix,
        "--resource-yml",
    ]

    # Run pipemake with the test command
    with unittest.mock.patch("sys.argv", test_cmd):
        with pytest.raises(Exception):
            main()


def test_pipemake_main_help_wo_error():
    os.environ["PM_SNAKEMAKE_DIR"] = "tests/files/snakemakeIO"

    # Assign the command line arguments
    test_cmd1 = [""]
    test_cmd2 = ["", "fastq-filter"]

    # Run pipemake with the test command
    with unittest.mock.patch("sys.argv", test_cmd1):
        with pytest.raises(SystemExit) as excinfo:
            main()
            excinfo.value.code == 2

    # Run pipemake with the test command
    with unittest.mock.patch("sys.argv", test_cmd2):
        with pytest.raises(SystemExit) as excinfo:
            main()
            excinfo.value.code == 2



import logging
import sys


def startLogger(log_filename=None, filemode="w"):
    # Close any old loggers
    for handler in logging.root.handlers[:]:
        handler.close()
        logging.root.removeHandler(handler)

    # Star the logger
    if log_filename:
        logging.basicConfig(
            filename=log_filename,
            filemode=filemode,
            level="INFO",
            format="%(asctime)s - %(levelname)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    else:
        logging.basicConfig(
            stream=sys.stdout,
            level="INFO",
            format="%(asctime)s - %(levelname)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )

    # Start logging to stdout
    stdout_log = logging.StreamHandler()

    # Assign the stdout logging level
    stdout_log.setLevel(logging.WARNING)

    # Define the stdout format
    console_format = logging.Formatter("%(funcName)s - %(levelname)s: %(message)s")

    # Assign the format
    stdout_log.setFormatter(console_format)

    # Add the stdout handler
    logging.getLogger("").addHandler(stdout_log)

    # Update the exception handler to log the exception
    def expHandler(etype, val, tb):
        # Log the error
        logging.error("%s" % (val), exc_info=(etype, val, tb))

    # Update the exception hook
    sys.excepthook = expHandler


def logArgDict(arg_dict, print_undefined=False, omit=[]):
    # Loop the arguments
    for arg, value in arg_dict.items():
        # Skip arg if in omit list
        if arg in omit:
            continue

        # Report only defined arguments, unless print_undefined is True
        if value is None and not print_undefined:
            continue

        # Log the argument
        logging.info("Argument %s: %s" % (arg, value))



import pytest
import os
import tempfile
import itertools

from pipemake.fileIO import checkIfGzipped, FileIO, TableIO


@pytest.mark.parametrize(
    "filename, expected",
    [
        ("tests/files/fileIO/bgzip.gz", True),
        ("tests/files/fileIO/gzip.gz", True),
        ("tests/files/fileIO/text.txt", False),
    ],
)
def test_checkIfGzipped(filename, expected):
    assert checkIfGzipped(filename) == expected


@pytest.mark.parametrize(
    "filename",
    ["tests/files/fileIO/test1_R1.fq", "tests/files/fileIO/test1_R2.fq"],
)
def test_FileIO_exists_wo_error(filename):
    try:
        FileIO.create(filename)
    except IOError:
        pytest.fail("FileIO.create() raised an error when opening a valid file.")


@pytest.mark.parametrize(
    "filename",
    ["tests/files/fileIO/test2_R1.fq", "tests/files/fileIO/test2_R2.fq"],
)
def test_FileIO_exists_w_error(filename):
    with pytest.raises(IOError):
        FileIO.create(filename)


@pytest.mark.parametrize(
    "filename, is_gzipped, standardized_filename, gzipped, copy_method",
    [
        ("tests/files/fileIO/test1_R1.fq", False, "test1_R1.fq", True, "symbolic_link"),
        (
            "tests/files/fileIO/test1_R1.fq.gz",
            True,
            "test1_R1.fq",
            True,
            "symbolic_link",
        ),
        (
            "tests/files/fileIO/test1_R1.fq",
            False,
            "test1_R1.fq",
            False,
            "symbolic_link",
        ),
        (
            "tests/files/fileIO/test1_R1.fq.gz",
            True,
            "test1_R1.fq",
            False,
            "symbolic_link",
        ),
        ("tests/files/fileIO/test1_R1.fq", False, "test1_R1.fq", None, "symbolic_link"),
        (
            "tests/files/fileIO/test1_R1.fq.gz",
            True,
            "test1_R1.fq",
            None,
            "symbolic_link",
        ),
        ("tests/files/fileIO/test1_R1.fq.gz", True, "test1_R1.fq", True, "copy"),
        ("tests/files/fileIO/test1_R1.fq", False, "test1_R1.fq", True, "copy"),
        ("tests/files/fileIO/test1_R1.fq", False, "test1_R1.fq", False, "copy"),
    ],
)
def test_FileIO_standardize(
    filename, is_gzipped, standardized_filename, gzipped, copy_method
):
    test_dir = tempfile.mkdtemp()
    test_File = FileIO.create(filename)
    test_File.standardize(
        standardized_filename,
        out_dir=test_dir,
        gzipped=gzipped,
        copy_method=copy_method,
    )

    # Test the copy method
    if copy_method == "copy":
        assert os.path.isfile(os.path.join(test_dir, standardized_filename))
        return

    # Test the symbolic link methods
    if is_gzipped is True and gzipped is True:
        assert os.path.islink(os.path.join(test_dir, standardized_filename))

    elif is_gzipped is False and gzipped is True:
        assert not os.path.islink(os.path.join(test_dir, standardized_filename))
        assert os.path.isfile(os.path.join(test_dir, standardized_filename))

    elif is_gzipped is False and gzipped is False:
        assert os.path.islink(os.path.join(test_dir, standardized_filename))

    elif is_gzipped is True and gzipped is False:
        assert not os.path.islink(os.path.join(test_dir, standardized_filename))
        assert os.path.isfile(os.path.join(test_dir, standardized_filename))

    elif is_gzipped is True and gzipped is None:
        assert os.path.islink(os.path.join(test_dir, standardized_filename))

    elif is_gzipped is False and gzipped is None:
        assert os.path.islink(os.path.join(test_dir, standardized_filename))


@pytest.mark.parametrize(
    "filename, copy_method",
    [
        ("tests/files/fileIO/test1_R1.fq", "symbolic_link"),
        ("tests/files/fileIO/test1_R1.fq", "copy"),
    ],
)
def test_FileIO_returnPaths(filename, copy_method):
    test_File = FileIO.create(filename)
    if copy_method == "copy":
        assert test_File.returnPaths(copy_method) == []
    else:
        assert test_File.returnPaths(copy_method) == [
            os.path.dirname(test_File.filename)
        ]


@pytest.mark.parametrize(
    "filename",
    ["tests/files/fileIO/test_table2.tsv"],
)
def test_TableIO_no_table_error(filename):
    with pytest.raises(IOError):
        TableIO.fromFilenameStr(filename)


@pytest.mark.parametrize(
    "filename",
    ["tests/files/fileIO/test_table.tsv"],
)
def test_TableIO_fromFilenameStr(filename):
    test_seqtable = TableIO.fromFilenameStr(filename, sample_column="samples")
    assert "samples" in test_seqtable.samples
    assert set(test_seqtable.samples["samples"]) == set(["test1", "test2"])
    assert "reads" in test_seqtable.samples
    assert set(test_seqtable.samples["reads"]) == set(["R1", "R2"])
    assert test_seqtable._sample_column == "samples"
    assert test_seqtable._file_columns == {"reads"}
    assert test_seqtable._table_columns == {"samples", "reads"}


@pytest.mark.parametrize(
    "filename",
    ["tests/files/fileIO/fasta_table.tsv"],
)
def test_TableIO_unique_column(filename):
    test_dir = tempfile.mkdtemp()
    test_seqtable = TableIO.fromFilenameStr(filename, sample_column="genomes")
    assert "genomes" in test_seqtable.samples
    assert set(test_seqtable.samples["genomes"]) == set(["genome1", "genome2"])
    assert test_seqtable._sample_column == "genomes"
    assert test_seqtable._file_columns == {"filename"}
    assert test_seqtable._table_columns == {"genomes"}

    test_seqtable.standardizedFiles(
        "{genomes}.fa", out_dir=test_dir, copy_method="symbolic_link"
    )
    assert os.path.islink(os.path.join(test_dir, "genome1.fa"))
    assert os.path.islink(os.path.join(test_dir, "genome2.fa"))

    for genome in ["genome1", "genome2"]:
        with open(os.path.join(test_dir, f"{genome}.fa")) as genome_file:
            assert (
                genome_file.read()
                == f">{genome.capitalize()}.1\nATCG\n>{genome.capitalize()}.2\nATCG"
            )


@pytest.mark.parametrize(
    "filename",
    ["tests/files/fileIO/id_table.tsv"],
)
def test_TableIO_no_file_column(filename):
    test_seqtable = TableIO.fromFilenameStr(filename, sample_column="SRA")
    assert "SRA" in test_seqtable.samples
    assert set(test_seqtable.samples["SRA"]) == set(
        ["SRR000001", "SRR000002", "SRR000003", "SRR000004"]
    )
    assert test_seqtable._sample_column == "SRA"
    assert test_seqtable._file_columns == set()
    assert test_seqtable._table_columns == {"SRA"}


@pytest.mark.parametrize(
    "filename, copy_method",
    [
        ("tests/files/fileIO/test_table.tsv", "symbolic_link"),
        ("tests/files/fileIO/test_table.tsv", "copy"),
    ],
)
def test_TableIO_standardizedFiles(filename, copy_method):
    test_dir = tempfile.mkdtemp()
    test_seqtable = TableIO.fromFilenameStr(filename, sample_column="samples")
    test_seqtable.standardizedFiles(
        "{samples}_{reads}.test.fq.gz", out_dir=test_dir, copy_method=copy_method
    )

    for samples, reads in list(itertools.product(["test1", "test2"], ["R1", "R2"])):
        standardized_filename = f"{samples}_{reads}.test.fq.gz"
        if copy_method == "symbolic_link":
            assert os.path.islink(os.path.join(test_dir, standardized_filename))
        elif copy_method == "copy":
            assert os.path.isfile(os.path.join(test_dir, standardized_filename))
        else:
            raise Exception(f"Unsupported copy method: {copy_method}")


@pytest.mark.parametrize(
    "filename, copy_method",
    [
        ("tests/files/fileIO/test_table.tsv", "symbolic_link"),
        ("tests/files/fileIO/test_table.tsv", "copy"),
    ],
)
def test_TableIO_returnPaths(filename, copy_method):
    test_seqtable = TableIO.fromFilenameStr(filename, sample_column="samples")
    if copy_method == "copy":
        assert test_seqtable.returnPaths(copy_method) == []
    else:
        assert test_seqtable.returnPaths(copy_method) == [
            os.path.abspath(os.path.dirname(filename))
        ]


@pytest.mark.parametrize(
    "filename",
    ["tests/files/fileIO/test_table.tsv"],
)
def test_TableIO_returnSamples(filename):
    test_seqtable = TableIO.fromFilenameStr(filename, sample_column="samples")
    assert test_seqtable.returnSamples()["samples"] == ["test1", "test2"]



import pytest
import os
import tempfile
import itertools

from pipemake.wildcardIO import WildcardIO


@pytest.mark.parametrize(
    "wildcard_str",
    [
        "tests/files/wildcardIO/{sample}_{read}.fq",
    ],
)
def test_WildcardIO_fromStr_w_error(wildcard_str):
    with pytest.raises(Exception):
        WildcardIO.fromStr(wildcard_str)


@pytest.mark.parametrize(
    "wildcard_str",
    ["tests/files/wildcardIO/{sample}_{read}.fq.gz"],
)
def test_WildcardIO_fromStr(wildcard_str):
    test_wildcard = WildcardIO.fromStr(wildcard_str)

    assert test_wildcard.wildcard_str == wildcard_str
    assert "sample" in test_wildcard.wildcard_dict
    assert "read" in test_wildcard.wildcard_dict

    sample_set = set(test_wildcard.wildcard_dict["sample"])
    assert sample_set == set(["test1", "test2"]) and len(sample_set) == 2

    read_set = set(test_wildcard.wildcard_dict["read"])
    assert read_set == set(["R1", "R2"]) and len(read_set) == 2


@pytest.mark.parametrize(
    "wildcard_str",
    ["tests/files/wildcardIO/{sample}_{read}.fq.gz"],
)
def test_WildcardIO_samples_w_error(wildcard_str):
    with pytest.raises(Exception):
        test_wildcard = WildcardIO.fromStr(wildcard_str)
        assert test_wildcard.samples == ["test1", "test2"]


@pytest.mark.parametrize(
    "wildcard_str, sample_wildcards",
    [("tests/files/wildcardIO/{sample}_{read}.fq.gz", ["sample", "read"])],
)
def test_WildcardIO_samples_wo_error(wildcard_str, sample_wildcards):
    test_wildcard = WildcardIO.fromStr(wildcard_str, sample_wildcards=sample_wildcards)
    samples_dict = test_wildcard.samples
    assert "sample" in samples_dict
    assert set(samples_dict["sample"]) == set(["test1", "test2"])
    assert "read" in samples_dict
    assert set(samples_dict["read"]) == set(["R1", "R2"])


@pytest.mark.parametrize(
    "wildcard_str, sample_wildcard, copy_method",
    [
        ("tests/files/wildcardIO/{sample}_{read}.fq.gz", "sample", "symbolic_link"),
        ("tests/files/wildcardIO/{sample}_{read}.fq.gz", "sample", "copy"),
    ],
)
def test_WildcardIO_standardizedFiles(wildcard_str, sample_wildcard, copy_method):
    test_dir = tempfile.mkdtemp()
    test_wildcard = WildcardIO.fromStr(wildcard_str, sample_wildcard=sample_wildcard)
    test_wildcard.standardizedFiles(
        "{sample}_{read}.test.fq.gz", out_dir=test_dir, copy_method=copy_method
    )

    for sample, read in list(itertools.product(["test1", "test2"], ["R1", "R2"])):
        standardized_filename = f"{sample}_{read}.test.fq.gz"
        if copy_method == "symbolic_link":
            assert os.path.islink(os.path.join(test_dir, standardized_filename))
        elif copy_method == "copy":
            assert os.path.isfile(os.path.join(test_dir, standardized_filename))
        else:
            raise Exception(f"Unsupported copy method: {copy_method}")


@pytest.mark.parametrize(
    "wildcard_str, sample_wildcard, copy_method",
    [
        ("tests/files/wildcardIO/{sample}_{read}.fq.gz", "sample", "symbolic_link"),
        ("tests/files/wildcardIO/{sample}_{read}.fq.gz", "sample", "copy"),
    ],
)
def test_WildcardIO_returnPaths(wildcard_str, sample_wildcard, copy_method):
    test_wildcard = WildcardIO.fromStr(wildcard_str, sample_wildcard=sample_wildcard)
    if copy_method == "copy":
        assert test_wildcard.returnPaths(copy_method) == []
    else:
        assert test_wildcard.returnPaths(copy_method) == [
            os.path.abspath(os.path.dirname(wildcard_str))
        ]



import pytest
import os
import tempfile
import yaml
import filecmp

from pipemake.snakemakeIO import (
    SnakePipelineIO,
    SnakeFileIO,
    SnakeRuleIO,
)


@pytest.mark.parametrize(
    "job_prefix, pipeline_storage_dir, resource_yml, scale_threads, scale_mem, indent_style, overwrite",
    [
        ("test", "tests/files/snakemakeIO2", None, 0, 0, "\t", True),
        ("test", "tests/files/snakemakeIO", "None", 0, 0, "\t", True),
        ("test", "tests/files/snakemakeIO", None, "0", 0, "\t", True),
        ("test", "tests/files/snakemakeIO", None, 0, "0", "\t", True),
        ("test", "tests/files/snakemakeIO", None, 0, "0", "\t", True),
        ("test", "tests/files/snakemakeIO", None, 0, 0, ",", True),
        ("test", "tests/files/snakemakeIO", None, 0, 0, "\t", None),
    ],
)
def test_SnakePipelineIO_w_error(
    job_prefix,
    pipeline_storage_dir,
    resource_yml,
    scale_threads,
    scale_mem,
    indent_style,
    overwrite,
):
    test_dir = tempfile.mkdtemp()
    singularity_dir = tempfile.mkdtemp()

    snakemake_job_prefix = os.path.join(test_dir, job_prefix)

    # Test if the function raises an error
    with pytest.raises(Exception):
        SnakePipelineIO.open(
            snakemake_job_prefix=snakemake_job_prefix,
            pipeline_storage_dir=pipeline_storage_dir,
            pipeline_job_dir=test_dir,
            resource_yml=resource_yml,
            scale_threads=scale_threads,
            scale_mem=scale_mem,
            singularity_dir=singularity_dir,
            indent_style=indent_style,
            overwrite=overwrite,
        )


@pytest.mark.parametrize(
    "job_prefix, pipeline_storage_dir, resource_yml, scale_threads, scale_mem, indent_style, overwrite",
    [("test", "tests/files/snakemakeIO", None, 1.0, 1.0, "\t", True)],
)
def test_SnakePipelineIO_wo_error(
    job_prefix,
    pipeline_storage_dir,
    resource_yml,
    scale_threads,
    scale_mem,
    indent_style,
    overwrite,
):
    test_dir = tempfile.mkdtemp()
    singularity_dir = tempfile.mkdtemp()

    snakemake_job_prefix = os.path.join(test_dir, job_prefix)

    # Create a test pipeline
    test_pipeline = SnakePipelineIO.open(
        snakemake_job_prefix=snakemake_job_prefix,
        workflow_prefix=snakemake_job_prefix,
        pipeline_storage_dir=pipeline_storage_dir,
        pipeline_job_dir=test_dir,
        resource_yml=resource_yml,
        scale_threads=scale_threads,
        scale_mem=scale_mem,
        singularity_dir=singularity_dir,
        indent_style=indent_style,
        overwrite=overwrite,
    )

    # Add a module to the pipeline
    test_pipeline.addModule("fastq_filter_fastp.smk")
    test_pipeline.addModule("test_script.smk")

    # Create the output directories
    unfiltered_fastq_dir = os.path.join(test_dir, "fastq")
    filtered_fastq_dir = os.path.join(test_dir, "fastq_filtered")

    # Write the pipeline config
    test_pipeline.writeConfig(
        {
            "min_length": 100,
            "samples": ["sample1", "sample2"],
            "unfiltered_fastq_dir": unfiltered_fastq_dir,
            "filtered_fastq_dir": filtered_fastq_dir,
            "workflow_prefix": snakemake_job_prefix,
        }
    )

    # Write the pipeline
    test_pipeline.writePipeline()

    # Build the singularity containers
    test_pipeline.buildSingularityContainers()

    test_pipeline.close()

    # Check if the pipeline was written
    assert os.path.exists(os.path.join(test_dir, "test.smk"))

    # Check if the module was written
    assert os.path.exists(os.path.join(test_dir, "modules", "fastq_filter_fastp.smk"))

    # Check if the script was written
    assert os.path.exists(os.path.join(test_dir, "scripts", "test.py"))

    # Check the contents of the script file
    with open(os.path.join(test_dir, "scripts", "test.py"), "r") as f:
        test_script_content = f.read()
        assert 'print("Script Found")' in test_script_content

    test_module_path = os.path.join(
        test_pipeline._module_job_dir, "fastq_filter_fastp.smk"
    )

    # Check if a string is within the file
    with open(os.path.join(test_dir, "test.smk"), "r") as f:
        test_pipeline_content = f.read()
        assert (
            'rule all:\n\tinput:\n\t\texpand(\n\t\t\tos.path.join(\n\t\t\t\tconfig["paths"]["workflow_prefix"],\n\t\t\t\tconfig["paths"]["filtered_fastq_dir"],\n\t\t\t\t"{sample}.json",\n\t\t\t),\n\t\t\tsample=config["samples"],\n\t\t),'
            in test_pipeline_content
        )
        assert f'include: "{test_module_path}"' in test_pipeline_content

    # Check if the config file was written
    assert os.path.exists(os.path.join(test_dir, "test.yml"))

    # Check if the singularity containers were built
    assert os.path.isfile(os.path.join(singularity_dir, "sambamba.v1.0.1.sif"))

    # Read the config yaml
    with open(os.path.join(test_dir, "test.yml"), "r") as yaml_file:
        config_dict = yaml.safe_load(yaml_file)
        assert "samples" in config_dict
        assert ["sample1", "sample2"] == config_dict["samples"]
        assert "min_length" in config_dict
        assert 100 == config_dict["min_length"]
        assert "fastp_single_end" in config_dict["resources"]
        assert "mem_mb" in config_dict["resources"]["fastp_single_end"]
        assert 16000 == config_dict["resources"]["fastp_single_end"]["mem_mb"]
        assert "threads" in config_dict["resources"]["fastp_single_end"]
        assert 4 == config_dict["resources"]["fastp_single_end"]["threads"]
        assert "fastp_pair_end" in config_dict["resources"]
        assert "mem_mb" in config_dict["resources"]["fastp_pair_end"]
        assert 16000 == config_dict["resources"]["fastp_pair_end"]["mem_mb"]
        assert "threads" in config_dict["resources"]["fastp_pair_end"]
        assert 4 == config_dict["resources"]["fastp_pair_end"]["threads"]
        assert "unfiltered_fastq_dir" in config_dict["paths"]
        assert unfiltered_fastq_dir == config_dict["paths"]["unfiltered_fastq_dir"]
        assert "filtered_fastq_dir" in config_dict["paths"]
        assert filtered_fastq_dir == config_dict["paths"]["filtered_fastq_dir"]


@pytest.mark.parametrize(
    "smk_filename", [("tests/files/snakemakeIO/modules/fastq_filter_fastp2.smk")]
)
def test_SnakeFileIO_w_error(smk_filename):
    # Test if the function raises an error
    with pytest.raises(Exception):
        SnakeFileIO.open(smk_filename=smk_filename)


@pytest.mark.parametrize(
    "smk_filename", [("tests/files/snakemakeIO/modules/fastq_filter_fastp.smk")]
)
def test_SnakeFileIO_wo_error(smk_filename):
    # Create a test directory
    test_dir = tempfile.mkdtemp()

    # Create a test module
    test_module = SnakeFileIO.open(smk_filename=smk_filename, singularity_dir=test_dir)

    # Create a out filename
    out_filename = os.path.join(test_dir, "test_fastq_filter_fastp.smk")
    cmp_filename = os.path.join(test_dir, "cmp_fastq_filter_fastp.smk")

    # Generate the module file and confirm it exists
    test_module.write(out_filename)
    print(out_filename)
    assert os.path.exists(out_filename)

    with open(cmp_filename, "w") as cmp_file:
        with open(
            "tests/files/snakemakeIO/test_fastq_filter_fastp.smk", "r"
        ) as test_file:
            for test_line in test_file:
                if "PATH" in test_line:
                    cmp_file.write(test_line.format(PATH=os.path.abspath(test_dir)))
                else:
                    cmp_file.write(test_line)

    # Check if the module was created correctly
    assert filecmp.cmp(out_filename, cmp_filename)


@pytest.mark.parametrize(
    "rule_filename", [("tests/files/snakemakeIO/rules/fastp_pair_end.smk")]
)
def test_SnakeRuleIO_wo_error(rule_filename):
    # Read the rule and store it as a string
    with open(rule_filename, "r") as test_file:
        test_str = test_file.read()

    # Test if the function raises an error
    test_rule = SnakeRuleIO.read(rule_str=test_str, indent_style="    ")

    # Read the rule and store it as a string
    with open("tests/files/snakemakeIO/test_fastp_pair_end.smk", "r") as cmp_file:
        cmp_str = cmp_file.read()

    # Check if the rule was created correctly
    assert test_rule._rule_text == cmp_str
    assert test_rule.rule_name == "fastp_pair_end"
    assert test_rule._rule_resource_params == {"threads": 4, "mem_mb": 16000}
    assert test_rule._rule_config_params == {
        ("paths", "workflow_prefix"),
        ("paths", "unfiltered_fastq_dir"),
        ("paths", "filtered_fastq_dir"),
        ("min_length",),
    }


@pytest.mark.parametrize(
    "rule_filename", [("tests/files/snakemakeIO/rules/test_script.smk")]
)
def test_SnakeRuleIO_w_script(rule_filename):
    # Read the rule and store it as a string
    with open(rule_filename, "r") as test_file:
        test_str = test_file.read()

    # Test if the function raises an error
    test_rule = SnakeRuleIO.read(rule_str=test_str, indent_style="    ")

    assert test_rule._rule_script_files == ["test.py"]



# Basic Information
__name__ = "pipemake"
__version__ = "1.1.6"
__summary__ = "Pipemake: A pipeline creation tool using Snakemake"
__url__ = "https://github.com/kocherlab/pipemake"
__code__ = "https://github.com/kocherlab/pipemake"
__issue__ = "https://github.com/kocherlab/pipemake/issues"
__docs__ = "https://pipemake.readthedocs.io/en/latest/"
__license__ = "MIT"
__copyright__ = "2023"

# Author Information
__authors__ = "Andrew Webb, Scott Wolf, Ian Traniello, and Sarah Kocher"
__email__ = "19213578+aewebb80@users.noreply.github.com"



import os
import logging
import subprocess


def checkSingularity():
    try:
        subprocess.run(
            ["singularity", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
    except FileNotFoundError:
        raise FileNotFoundError("Singularity is not installed on this system")


class Singularity:
    def __init__(self, url="", singularity_path=""):
        # Check if singularity is installed
        checkSingularity()

        # Set the singularity attributes
        self._url = url
        self._image_filename = ""
        self._singularity_path = singularity_path

        # If the url is provided, process it
        if self._url:
            self.processURL()

    def processURL(self):
        # Set the image name and tag
        image_name, image_tag = self._url.rsplit("/", 1)[-1].split(":")

        # Set the image name
        image_filename = f"{image_name}.{image_tag}.sif"

        # Set the image path if the singularity path is provided
        if self._singularity_path:
            image_filename = os.path.join(self._singularity_path, image_filename)

        # Set the image filename
        self._image_filename = os.path.abspath(image_filename)

    def download(self):
        # Check if the image already exists
        if os.path.isfile(self._image_filename):
            logging.info(f"Image already exists at: {self._image_filename}")
            print(f"Image already exists at: {self._image_filename}")
            return self._image_filename

        # Check if the singularity path exists
        if self._singularity_path and not os.path.isdir(self._singularity_path):
            os.makedirs(self._singularity_path)

        logging.info(f"Downloading image: {self._url}")
        print(f"Downloading image: {self._url}")

        # Download the image using singularity
        singularity_process = subprocess.Popen(
            ["singularity", "pull", self._image_filename, self._url],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # Get the stdout and stderr
        _, singularity_stderr = singularity_process.communicate()

        # Check for errors
        if "FATAL" in singularity_stderr.decode():
            raise Exception(singularity_stderr.decode().strip())

        # Confirm the image was downloaded
        if not os.path.isfile(self._image_filename):
            raise Exception("Image was not downloaded")

        logging.info(f"Image created at: {self._image_filename}")
        print(f"Image created at: {self._image_filename}")

    def returnPath(self):
        return self._image_filename

    def updateContainer(self, str):
        return str.replace(self._url, self._image_filename)

    @classmethod
    def fromURL(cls, url, singularity_path=""):
        return cls(url=url, singularity_path=singularity_path)



#!/usr/bin/env python

import os
import shutil

from pipemake.logger import startLogger, logArgDict
from pipemake.parser import PipelineParser
from pipemake.snakemakeIO import SnakePipelineIO
from pipemake.pipelineIO import ConfigPipelinesIO


def main():
    # Assign the pipeline directory
    if os.environ.get("PM_SNAKEMAKE_DIR"):
        pipeline_storage_dir = os.environ.get("PM_SNAKEMAKE_DIR")
    else:
        pipeline_storage_dir = "pipelines"

    # Confirm the pipeline directory exists
    if not os.path.isdir(pipeline_storage_dir):
        raise Exception(f"Unable to find pipeline directory: {pipeline_storage_dir}")

    # Load the pipeline configs
    pipeline_configs = ConfigPipelinesIO.fromDirectory(pipeline_storage_dir)

    # Parse the aguments from the configs
    pipeline_parser = PipelineParser.create(pipeline_configs)
    pipeline_args = pipeline_parser.returnArgs()

    # Assign the pipeline directory to an environment variable, if found
    if os.environ.get("PM_SINGULARITY_DIR") and not pipeline_args["singularity_dir"]:
        pipeline_args["singularity_dir"] = os.environ.get("PM_SINGULARITY_DIR")

    # Update the pipeline args with the pipeline directory
    pipeline_args["pipeline_storage_dir"] = pipeline_storage_dir

    # Check that a pipeline was assigned
    if not pipeline_args["pipeline"]:
        raise Exception("No pipeline specified")

    # Check for a previous workflow directory

    if os.path.exists(pipeline_args["workflow_prefix"]) and pipeline_args["overwrite"]:
        shutil.rmtree(pipeline_args["workflow_prefix"])
    elif os.path.exists(pipeline_args["workflow_prefix"]):
        raise Exception(
            f"Workflow directory already exists: {pipeline_args['workflow_prefix']}"
        )

    # Assign the pipeline job directory
    pipeline_args["pipeline_job_dir"] = os.path.join(
        pipeline_args["workflow_prefix"], "pipemake"
    )

    # Check if the pipeline job directory should be updated
    if pipeline_args[
        "work_dir"
    ]:  # and not os.path.exists(pipeline_args['work_dir']): os.makedirs(pipeline_args['work_dir'])
        pipeline_args["pipeline_job_dir"] = os.path.join(
            pipeline_args["work_dir"], pipeline_args["pipeline_job_dir"]
        )

    # Create the pipeline job directory
    if not os.path.exists(pipeline_args["pipeline_job_dir"]):
        os.makedirs(pipeline_args["pipeline_job_dir"])

    # Start logger and log the arguments
    startLogger(os.path.join(pipeline_args["pipeline_job_dir"], "pipeline.log"))
    logArgDict(pipeline_args, omit=["pipeline_job_dir"])

    # Assign the pipeline config
    pipline_config = pipeline_configs[pipeline_args["pipeline"]]

    # Process the pipeline setup
    pipline_config.setupPipeline(pipeline_args)

    # Add the samples to the pipeline args
    pipeline_args.update(pipline_config.samples)

    # Create the snakemake pipeline
    snakemake_pipeline = SnakePipelineIO.open(**pipeline_args)

    # Add the snakemake modules to the pipeline
    for smkm_filename in pipline_config.snakefiles:
        snakemake_pipeline.addModule(smkm_filename)

    # Build the singularity containers
    snakemake_pipeline.buildSingularityContainers()

    # Create the snakemake config file
    snakemake_pipeline.writeConfig(pipeline_args)

    # Create the snakemake piepline file
    snakemake_pipeline.writePipeline()

    # Close the snakemake pipeline
    snakemake_pipeline.close()

    # Print the singularity help message
    pipline_config.helpMessage(pipeline_args)


if __name__ == "__main__":
    main()



import os
import pytest
import tempfile

from pipemake.logger import *


@pytest.mark.parametrize(
    "log_filename",
    ["test.log"],
)
def test_logger_startLogger(log_filename):
    test_dir = tempfile.mkdtemp()

    # Assign the log file path
    log_file_path = os.path.join(test_dir, log_filename)

    # Start the logger
    startLogger(log_filename=log_file_path)

    # Log a message
    logging.info("This is a test message")

    # Check if the log file was created
    assert os.path.exists(log_file_path)


@pytest.mark.parametrize(
    "log_filename",
    ["test.log"],
)
def test_logger_logArgDict(log_filename):
    test_dir = tempfile.mkdtemp()

    # Assign the log file path
    test_dict = {"arg1": "value1", "arg2": None, "arg3": "value3"}

    # Assign the log file path
    log_file_path = os.path.join(test_dir, log_filename)

    # Start the logger
    startLogger(log_filename=log_file_path)

    # Log the argument dictionary
    logArgDict(test_dict)

    # Open the log file
    with open(log_file_path, "r") as log_file:
        # Read the log file
        log_text = log_file.read()

        # Check if the log file contains the expected text
        assert "Argument arg1: value1" in log_text
        assert "Argument arg2: None" not in log_text
        assert "Argument arg3: value3" in log_text



import os
import pytest
import tempfile
import itertools

from pipemake.processIO import ProcessIO, standardizeInput, returnPaths, returnSamples


@pytest.mark.parametrize(
    "wildcard_str",
    [
        "tests/files/wildcardIO/{samples}_{reads}.fq",
    ],
)
def test_processIO_fromWildcardStr_w_error(wildcard_str):
    with pytest.raises(Exception):
        ProcessIO.fromWildcardStr(wildcard_str)


@pytest.mark.parametrize(
    "wildcard_str",
    ["tests/files/wildcardIO/{samples}_{reads}.fq.gz"],
)
def test_processIO_fromWildcardStr_wo_error(wildcard_str):
    ProcessIO.fromWildcardStr(wildcard_str)


@pytest.mark.parametrize(
    "wildcard_str, standardized_wildcard",
    [
        (
            "tests/files/wildcardIO/{samples}_{reads}.fq.gz",
            "{samples}_{reads}.test.fq.gz",
        )
    ],
)
def test_processIO_fromWildcardStr_standardizeInput(
    wildcard_str, standardized_wildcard
):
    test_dir = tempfile.mkdtemp()
    standardizeInput(
        method="wildcard-str",
        args={
            "wildcard_str": wildcard_str,
            "standardized_filename": standardized_wildcard,
            "out_dir": test_dir,
        },
    )

    for sample, read in list(itertools.product(["test1", "test2"], ["R1", "R2"])):
        standardized_filename = f"{sample}_{read}.test.fq.gz"
        assert os.path.isfile(os.path.join(test_dir, standardized_filename))


@pytest.mark.parametrize(
    "wildcard_str",
    [("tests/files/wildcardIO/{samples}_{reads}.fq.gz")],
)
def test_processIO_fromWildcardStr(wildcard_str):
    assert returnPaths(method="wildcard-str", args={"wildcard_str": wildcard_str}) == [
        os.path.abspath(os.path.dirname(wildcard_str))
    ]

    samples_dict = returnSamples(
        method="wildcard-str",
        args={"wildcard_str": wildcard_str, "sample_wildcards": ["samples"]},
    )
    assert "samples" in samples_dict
    assert set(samples_dict["samples"]) == set(["test1", "test2"])


@pytest.mark.parametrize(
    "file_str",
    ["tests/files/fileIO/test2_R1.fq", "tests/files/fileIO/test2_R2.fq"],
)
def test_processIO_fromFileStr_w_error(file_str):
    with pytest.raises(IOError):
        ProcessIO.fromFileStr(file_str)


@pytest.mark.parametrize(
    "file_str",
    ["tests/files/fileIO/test1_R1.fq", "tests/files/fileIO/test1_R2.fq"],
)
def test_processIO_fromFileStr_wo_error(file_str):
    ProcessIO.fromFileStr(file_str)


@pytest.mark.parametrize(
    "file_str, standardized_filename",
    [
        ("tests/files/fileIO/test1_R1.fq", "test1_R1.test.fq.gz"),
        ("tests/files/fileIO/test1_R2.fq", "test1_R2.test.fq.gz"),
    ],
)
def test_processIO_fromFileStr_standardizeInput_wo_error(
    file_str, standardized_filename
):
    test_dir = tempfile.mkdtemp()
    standardizeInput(
        method="file-str",
        args={
            "input_filename": file_str,
            "standardized_filename": standardized_filename,
            "out_dir": test_dir,
        },
    )
    assert os.path.isfile(os.path.join(test_dir, standardized_filename))


@pytest.mark.parametrize(
    "file_str, standardized_filename",
    [
        ("tests/files/fileIO/test1_R1.fq", "test1_R1.test.fq.gz"),
        ("tests/files/fileIO/test1_R2.fq", "test1_R2.test.fq.gz"),
    ],
)
def test_processIO_fromFileStr_standardizeInput_w_error(
    file_str, standardized_filename
):
    assert returnPaths(
        method="file-str",
        args={
            "input_filename": file_str,
            "standardized_filename": standardized_filename,
        },
    ) == [os.path.abspath(os.path.dirname(file_str))]
    with pytest.raises(Exception):
        returnSamples(method="file-str", args={"input_filename": file_str})


@pytest.mark.parametrize(
    "table_file",
    ["tests/files/fileIO/test_table2.tsv"],
)
def test_processIO_fromTableFile_w_error(table_file):
    with pytest.raises(IOError):
        ProcessIO.fromTableFile(table_file)


@pytest.mark.parametrize(
    "table_file",
    ["tests/files/fileIO/test_table.tsv"],
)
def test_processIO_fromTableFile_wo_error(table_file):
    ProcessIO.fromTableFile(table_file, sample_column="samples")


@pytest.mark.parametrize(
    "table_file, standardized_wildcard",
    [
        ("tests/files/fileIO/test_table.tsv", "{samples}_{reads}.test.fq.gz"),
    ],
)
def test_processIO_fromTableFile_standardizeInput(table_file, standardized_wildcard):
    test_dir = tempfile.mkdtemp()
    standardizeInput(
        method="table-file",
        args={
            "table_filename": table_file,
            "standardized_filename": standardized_wildcard,
            "out_dir": test_dir,
            "sample_column": "samples",
        },
    )

    for sample, read in list(itertools.product(["test1", "test2"], ["R1", "R2"])):
        standardized_filename = f"{sample}_{read}.test.fq.gz"
        assert os.path.isfile(os.path.join(test_dir, standardized_filename))


@pytest.mark.parametrize(
    "table_file",
    [
        ("tests/files/fileIO/test_table.tsv"),
    ],
)
def test_processIO_fromTableFile(table_file):
    assert returnPaths(
        method="table-file",
        args={"table_filename": table_file, "sample_column": "samples"},
    ) == [os.path.abspath(os.path.dirname(table_file))]
    samples_dict = returnSamples(
        method="table-file",
        args={"table_filename": table_file, "sample_column": "samples"},
    )
    assert "samples" in samples_dict
    assert set(samples_dict["samples"]) == set(["test1", "test2"])
    assert "reads" in samples_dict
    assert set(samples_dict["reads"]) == set(["R1", "R2"])



from pipemake.fileIO import FileIO, TableIO
from pipemake.wildcardIO import WildcardIO
from pipemake.pathIO import DirIO


class ProcessIO:
    def __init__(
        self, processIO, standardize_func=None, standardize_type=None, **kwargs
    ):
        self.processIO = processIO
        self.standardize_call = getattr(self.processIO, standardize_func)
        self.standardize_type = standardize_type

    @classmethod
    def fromWildcardStr(cls, wildcard_str="", **kwargs):
        return cls(
            WildcardIO.fromStr(wildcard_str, **kwargs),
            standardize_func="standardizedFiles",
            standardize_type="file",
        )

    @classmethod
    def fromTableFile(cls, table_filename="", **kwargs):
        return cls(
            TableIO.fromFilenameStr(table_filename, **kwargs),
            standardize_func="standardizedFiles",
            standardize_type="file",
        )

    @classmethod
    def fromFileStr(cls, input_filename="", **kwargs):
        return cls(
            FileIO.create(input_filename, **kwargs),
            standardize_func="standardize",
            standardize_type="file",
        )

    @classmethod
    def fromDirStr(cls, path_name="", **kwargs):
        return cls(
            DirIO.create(path_name, **kwargs),
            standardize_func="standardize",
            standardize_type="dir",
        )

    def standardize(self, **kwargs):
        if self.standardize_type == "file":
            self._standardize_file(**kwargs)
        elif self.standardize_type == "dir":
            self._standardize_dir(**kwargs)
        else:
            raise Exception(f"Standardize type not recognized: {self.standardize_type}")

    def _standardize_file(self, standardized_filename="", **kwargs):
        self.standardize_call(standardized_filename, **kwargs)

    def _standardize_dir(self, standardized_directory="", **kwargs):
        self.standardize_call(standardized_directory, **kwargs)

    def returnSamples(self, **kwargs):
        return self.processIO.returnSamples()

    def returnPaths(self, **kwargs):
        return self.processIO.returnPaths(**kwargs)


def standardizeInput(method="", args={}):
    # Create the standardization call
    if method == "wildcard-str":
        standardize_input_call = ProcessIO.fromWildcardStr(**args)
    elif method == "table-file":
        standardize_input_call = ProcessIO.fromTableFile(**args)
    elif method == "file-str":
        standardize_input_call = ProcessIO.fromFileStr(**args)
    elif method == "dir-str":
        standardize_input_call = ProcessIO.fromDirStr(**args)
    else:
        raise Exception(f"No standardization method given for: {method}")

    # Standardize the input
    standardize_input_call.standardize(**args)


def returnPaths(method="", args={}):
    # Create the standardization call
    if method == "wildcard-str":
        return_path_call = ProcessIO.fromWildcardStr(**args)
    elif method == "table-file":
        return_path_call = ProcessIO.fromTableFile(**args)
    elif method == "file-str":
        return_path_call = ProcessIO.fromFileStr(**args)
    elif method == "dir-str":
        return_path_call = ProcessIO.fromDirStr(**args)
    else:
        raise Exception(f"No standardization method given for: {method}")

    return return_path_call.returnPaths(**args)


def returnSamples(method="", args={}):
    # Create the return samples call
    if method == "wildcard-str":
        return_samples_call = ProcessIO.fromWildcardStr(**args)
    elif method == "table-file":
        return_samples_call = ProcessIO.fromTableFile(**args)
    else:
        raise Exception(f"No sample method given for: {method}")

    return return_samples_call.returnSamples()



import pytest

from pipemake.pipelineIO import ConfigPipelinesIO, ConfigPipelineIO


@pytest.mark.parametrize(
    "config_path",
    ["tests/files/snakemakeIO/"],
)
def test_ConfigPipelinesIO_fromDirectory_wo_error(config_path):
    # Read in the pipeline configs
    configs = ConfigPipelinesIO.fromDirectory(config_path)

    # Check the pipeline configs
    assert isinstance(configs, ConfigPipelinesIO)
    assert len(configs) == 1
    assert "fastq-filter" in configs
    assert isinstance(configs["fastq-filter"], ConfigPipelineIO)


@pytest.mark.parametrize(
    "config_path",
    ["tests/files/snakemakeIO/configs"],
)
def test_ConfigPipelinesIO_fromDirectory_w_error(config_path):
    with pytest.raises(Exception):
        ConfigPipelinesIO.fromDirectory(config_path)


@pytest.mark.parametrize(
    "config_yaml",
    ["tests/files/snakemakeIO/configs/fastq_filter.yml"],
)
def test_ConfigPipelineIO_fromYAML_wo_error(config_yaml):
    # Read in the pipeline configs
    config = ConfigPipelineIO.fromYAML(config_yaml)

    # Check the pipeline configs
    assert isinstance(config, ConfigPipelineIO)
    assert config.name == "fastq-filter"
    assert config.version == 1.0


@pytest.mark.parametrize(
    "config_yaml",
    ["tests/files/snakemakeIO/configs/fastq_filter.yaml"],
)
def test_ConfigPipelineIO_fromYAML_w_error(config_yaml):
    with pytest.raises(Exception):
        ConfigPipelineIO.fromYAML(config_yaml)


@pytest.mark.parametrize(
    "config_yaml",
    ["tests/files/snakemakeIO/configs/fastq_filter.yml"],
)
def test_ConfigPipelineIO_fromYAML_parser_dict(config_yaml):
    # Read in the pipeline configs
    config = ConfigPipelineIO.fromYAML(config_yaml)

    assert "help" in config.parser_dict
    assert "arg-groups" in config.parser_dict
    assert "basic" in config.parser_dict["arg-groups"]
    assert "paths" in config.parser_dict["arg-groups"]
    assert "args" in config.parser_dict["arg-groups"]["basic"]
    assert "args" in config.parser_dict["arg-groups"]["paths"]
    assert "mutually-exclusive-args" in config.parser_dict["arg-groups"]["basic"]
    assert (
        "input-parser"
        in config.parser_dict["arg-groups"]["basic"]["mutually-exclusive-args"]
    )
    assert "fastq-wildcard" in config.parser_dict["arg-groups"]["basic"]["args"]
    assert "type" in config.parser_dict["arg-groups"]["basic"]["args"]["fastq-wildcard"]
    assert "help" in config.parser_dict["arg-groups"]["basic"]["args"]["fastq-wildcard"]
    assert (
        "mutually-exclusive"
        in config.parser_dict["arg-groups"]["basic"]["args"]["fastq-wildcard"]
    )


@pytest.mark.parametrize(
    "config_yaml",
    ["tests/files/snakemakeIO/configs/fastq_filter.yml"],
)
def test_ConfigPipelineIO_fromYAML_setup_dict(config_yaml):
    # Read in the pipeline configs
    config = ConfigPipelineIO.fromYAML(config_yaml)

    assert "fastq_input" in config._setup_dict
    assert "wildcard-method" in config._setup_dict["fastq_input"]
    assert "input" in config._setup_dict["fastq_input"]["wildcard-method"]
    assert "args" in config._setup_dict["fastq_input"]["wildcard-method"]["input"]
    assert set(
        [
            "workflow_prefix",
            "fastq-wildcard",
            "fastq-standardized-wildcard",
            "unfiltered-fastq-dir",
        ]
    ) == set(config._setup_dict["fastq_input"]["wildcard-method"]["input"]["args"])
    assert "standardize" in config._setup_dict["fastq_input"]["wildcard-method"]
    assert (
        "method" in config._setup_dict["fastq_input"]["wildcard-method"]["standardize"]
    )
    assert "args" in config._setup_dict["fastq_input"]["wildcard-method"]["standardize"]
    assert (
        "out_dir"
        in config._setup_dict["fastq_input"]["wildcard-method"]["standardize"]["args"]
    )
    assert (
        "gzipped"
        in config._setup_dict["fastq_input"]["wildcard-method"]["standardize"]["args"]
    )


@pytest.mark.parametrize(
    "config_yaml",
    ["tests/files/snakemakeIO/configs/fastq_filter.yml"],
)
def test_ConfigPipelineIO_fromYAML_snakefiles(config_yaml):
    # Read in the pipeline configs
    config = ConfigPipelineIO.fromYAML(config_yaml)

    assert "fastq_filter_fastp" in config._snakefiles



import os
import shutil
import logging


class DirIO:
    def __init__(self, path_name, **kwargs):
        # Confirm the directory exists
        if not os.path.isdir(path_name):
            raise IOError(
                f"Unable to locate: {path_name}. Please confirm the directory exists."
            )

        # Convert the directory to an absolute path
        abs_path_name = os.path.abspath(path_name)

        # Confirm the directory using the absolute path
        if not os.path.isdir(abs_path_name):
            raise IOError(
                f"Unable to locate: {abs_path_name}. Please confirm the directory exists."
            )

        # Assign the basic arguments
        self.path_name = abs_path_name

        # Stores path if using a link
        self.link_path = ""

    @classmethod
    def create(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def standardize(
        self,
        standardized_directory,
        out_dir="",
        workflow_prefix="",
        work_dir="",
        copy_method="symbolic_link",
        **kwargs,
    ):
        # Assign the destination directory
        dest_directory = standardized_directory

        # Create path as needed
        if out_dir:
            dest_directory = os.path.join(out_dir, dest_directory)
        if workflow_prefix:
            dest_directory = os.path.join(workflow_prefix, dest_directory)
        if work_dir:
            dest_directory = os.path.join(work_dir, dest_directory)

        # Create the output directory, if needed
        if not os.path.exists(os.path.dirname(dest_directory)):
            os.makedirs(os.path.dirname(dest_directory))

        # Copy the directory
        if copy_method == "symbolic_link":
            os.symlink(self.path_name, dest_directory)
        elif copy_method == "copy":
            shutil.copytree(self.path_name, dest_directory)
        else:
            raise Exception(f"Unknown copy method: {copy_method}")

        logging.info(
            f"Standardized: {self.path_name} to {dest_directory}. Method: {copy_method}"
        )

    def returnPaths(self, copy_method="symbolic_link", **kwargs):
        if copy_method == "copy":
            return []
        elif copy_method == "symbolic_link":
            return [self.path_name]
        else:
            raise Exception(f"Unsupported copy method: {copy_method}")



import os
import copy
import yaml
import logging

from pipemake.processIO import standardizeInput, returnSamples, returnPaths


class ConfigPipelinesIO(dict):
    def __init__(self, *arg, **kwargs):
        super(ConfigPipelinesIO, self).__init__(*arg, **kwargs)

    @classmethod
    def fromDirectory(cls, directory):
        # Check the config directory exists
        config_dir = os.path.join(directory, "configs")
        if not os.path.isdir(config_dir):
            raise Exception(f"Unable to find pipeline directory: {config_dir}")

        # Create dict to store the pipeline configs
        pipeline_configs = {}

        # Loop the configs
        for config_filename in os.listdir(config_dir):
            config_file = os.path.join(config_dir, config_filename)

            # Confirm the path is a file
            if not os.path.isfile(config_file):
                continue

            # Load the config file
            pipeline_config = ConfigPipelineIO.fromYAML(config_file)

            # Store the pipeline config
            pipeline_configs[pipeline_config.name] = pipeline_config

        return cls(**pipeline_configs)


class ConfigPipelineIO:
    def __init__(self, config_dict):
        # Check if the pipeline name is provided
        if "pipeline" not in config_dict:
            raise Exception("No pipeline name provided in the configuration file")

        # Set the pipeline name
        self.name = config_dict["pipeline"]
        config_dict.pop("pipeline")

        # Check if the pipeline version is provided
        if "version" not in config_dict:
            raise Exception("No pipeline version provided in the configuration file")

        # Set the pipeline version
        self.version = config_dict["version"]
        config_dict.pop("version")

        # Check if the pipeline parser is provided
        if "parser" not in config_dict:
            raise Exception("No parser provided in the configuration file")

        # Set the pipeline parser
        self.parser_dict = config_dict["parser"]
        config_dict.pop("parser")

        # Check if the pipeline setup is provided
        if "setup" not in config_dict:
            logging.warning("No setup provided in the configuration file")

        # Set the pipeline setup
        self._setup_dict = config_dict["setup"]
        config_dict.pop("setup")

        # Check if the pipeline snakefiles are provided
        if "snakefiles" not in config_dict:
            raise Exception("No snakefiles provided in the configuration file")

        # Set the pipeline snakefiles
        self._snakefiles = set(config_dict["snakefiles"])
        config_dict.pop("snakefiles")

        # Record unused configuration parameters
        for key in config_dict:
            logging.info(f"Unused configuration parameter: {key}")

        # Assign the other configuration arguments
        self._singularity_bindings = set()
        self.samples = {}

    @property
    def snakefiles(self):
        return list(self._snakefiles)

    @property
    def singularity_bindings(self):
        return list(self._singularity_bindings)

    @classmethod
    def fromYAML(cls, config_filename):
        with open(config_filename, "r") as config_file:
            config_dict = yaml.safe_load(config_file)
        return cls(config_dict)

    def helpMessage(self, pipeline_args):
        print(
            "If running the pipeline using singularity containers, please use the following command:"
        )
        logging.info(
            "If running the pipeline using singularity containers, please use the following command:"
        )

        if self.singularity_bindings:
            print(
                f"snakemake -s {pipeline_args['workflow_prefix']}.smk --use-singularity --singularity-args '--bind {','.join(self.singularity_bindings)}'"
            )
            logging.info(
                f"snakemake -s {pipeline_args['workflow_prefix']}.smk --use-singularity --singularity-args '--bind {','.join(self.singularity_bindings)}'"
            )
        else:
            print(
                f"snakemake -s {pipeline_args['workflow_prefix']}.smk --use-singularity"
            )
            logging.info(
                f"snakemake -s {pipeline_args['workflow_prefix']}.smk --use-singularity"
            )

    def setupPipeline(self, pipeline_args):
        # Confirm the workflow prefix is specified
        if "workflow_prefix" not in pipeline_args:
            raise Exception("Workflow prefix not found among pipeline arguments")

        for setup_name, setup_methods in self._setup_dict.items():
            # Create a string to store the assigned method
            assigned_method = ""

            for method_name, method_args in setup_methods.items():
                # Assign the input args
                input_args = method_args["input"]

                # Check for missing arguments
                for input_arg in input_args["args"]:
                    if input_arg.replace("-", "_") not in pipeline_args:
                        raise Exception(
                            f"Setup argument {input_arg} not found among pipeline argument"
                        )

                # Confirm expected args were specified
                method_missing_args = [
                    _a
                    for _a in input_args["args"]
                    if not pipeline_args[_a.replace("-", "_")]
                ]

                # Skip if missing arguments
                if method_missing_args:
                    continue

                # Check if the method was already assigned and report an error if so
                if assigned_method:
                    raise Exception(
                        f"Setup {setup_name} already assigned to {assigned_method}, cannot assign {method_name}"
                    )
                else:
                    assigned_method = method_name

                if "standardize" in method_args:
                    # Create a dict of the standardize args
                    std_args = copy.deepcopy(method_args["standardize"])

                    # Add the workflow prefix to the args
                    std_args["args"]["workflow_prefix"] = "{workflow_prefix}"

                    # Check if the work_dir is specified
                    if "work_dir" in pipeline_args and pipeline_args["work_dir"]:
                        std_args["args"]["work_dir"] = "{work_dir}"

                    # Update the arguments
                    for std_arg, arg_params in std_args["args"].items():
                        if not isinstance(arg_params, str):
                            continue
                        std_args["args"][std_arg] = arg_params.replace("-", "_").format(
                            **pipeline_args
                        )

                    # Standardize the input
                    standardizeInput(**std_args)

                    # Assign the method paths
                    method_paths = returnPaths(**std_args)

                    # Check for method paths, and update the singularity bindings if found
                    if len(method_paths) > 0:
                        self._singularity_bindings.update(method_paths)

                if "samples" in method_args:
                    # Create a dict of the standardize args
                    samples_args = copy.deepcopy(method_args["samples"])

                    # Update the arguments
                    for samples_arg, arg_params in samples_args["args"].items():
                        if not isinstance(arg_params, str):
                            continue
                        samples_args["args"][samples_arg] = arg_params.replace(
                            "-", "_"
                        ).format(**pipeline_args)

                    # Assign the samples from the method
                    method_samples = returnSamples(**samples_args)

                    # Confirm the samples are not already assigned
                    if method_samples and len(self.samples) > 0:
                        raise Exception("Samples already assigned")

                    # Store the samples
                    self.samples = method_samples

                if "snakefiles" in method_args:
                    # Add method snakefiles to the pipeline
                    self._snakefiles.update(method_args["snakefiles"])



import os
import re
import sys
import random
import string
import logging
import argparse
import datetime

from pydoc import locate


def jobRandomString(num_chars=4):
    global random_string

    # Create a random string, if not done already
    if not random_string:
        random_string = "".join(
            random.choices(string.digits + string.ascii_uppercase, k=num_chars)
        )

    return random_string


def jobTimeStamp():
    global time_stamp

    # Create a time stamp, if not done already
    if not time_stamp:
        time_stamp = datetime.datetime.now().strftime("%Y-%m-%d")

    return time_stamp


def confirmDir():
    """Custom action to confirm directory exists"""

    class customAction(argparse.Action):
        def __call__(self, parser, args, value, option_string=None):
            if not os.path.isdir(value):
                raise IOError(f"Unable to find directory: {value}")
            setattr(args, self.dest, value)

    return customAction


def confirmFile():
    """Custom action to confirm file exists"""

    class customAction(argparse.Action):
        def __call__(self, parser, args, value, option_string=None):
            if not os.path.isfile(value):
                raise IOError(f"Unable to find file: {value}")
            setattr(args, self.dest, value)

    return customAction


class MyParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write("error: %s\n\n" % message)
        self.print_help()
        sys.exit(2)


class SubcommandHelpFormatter(argparse.RawDescriptionHelpFormatter):
    def _format_action(self, action):
        parts = super(argparse.RawDescriptionHelpFormatter, self)._format_action(action)
        if action.nargs == argparse.PARSER:
            parts = "\n".join(parts.split("\n")[1:])
        return parts


class PipelineParser:
    def __init__(self, config_pipelines):
        # Create the pipeline selection parser
        self._pipeline_parser = MyParser(formatter_class=SubcommandHelpFormatter)
        self._pipeline_parser._positionals.title = (
            "Pipeline selection argument options (positional)"
        )

        # Create the pipeline subparsers
        self._pipeline_subparsers = self._pipeline_parser.add_subparsers(
            dest="pipeline", required=True
        )

        # Ceate a dictionary to store each subparser
        self._pipeline_subparsers_dict = {}

        for pipeline_name, pipeline_config in config_pipelines.items():
            # Assign the arguments for each pipeline
            self._assignPipeline(pipeline_name, pipeline_config)

    def _assignPipeline(self, pipeline_name, pipeline_config):
        # Assign the pipeline parser
        pipeline_parser_dict = pipeline_config.parser_dict

        # Create the subparser
        self._pipeline_subparsers_dict[pipeline_name] = (
            self._pipeline_subparsers.add_parser(
                pipeline_name,
                help=pipeline_parser_dict["help"] + f" (v{pipeline_config.version})",
                add_help=False,
            )
        )

        # Create the argument groups for the pipeline
        pipeline_arg_groups = {}

        # Add the required argument group
        pipeline_arg_groups["required"] = self._pipeline_subparsers_dict[
            pipeline_name
        ].add_argument_group(f"{pipeline_name} required arguments")

        # Loop the pipeline groups
        for pipeline_arg_group in pipeline_parser_dict["arg-groups"]:
            # Create the argument group, if not the basic subparsers
            if pipeline_arg_group == "basic":
                continue

            # Check if the group already exists, then raise an exception if it does
            if pipeline_arg_group in pipeline_arg_groups:
                raise Exception(
                    f"Cannot create group {pipeline_arg_group}. Group already exists."
                )

            # Create the argument group
            pipeline_arg_groups[pipeline_arg_group] = self._pipeline_subparsers_dict[
                pipeline_name
            ].add_argument_group(f"{pipeline_name} {pipeline_arg_group} arguments")

        # Add the optional argument group
        pipeline_arg_groups["optional"] = self._pipeline_subparsers_dict[
            pipeline_name
        ].add_argument_group(f"{pipeline_name} optional arguments")

        # Loop the pipeline argument groups
        for pipeline_arg_group, group_args in pipeline_parser_dict[
            "arg-groups"
        ].items():
            # Check if mutually exclusive arguments are present
            for mutually_exclusive_name, mutually_exclusive_args in self._yeildGroup(
                "mutually-exclusive-args", group_args
            ):
                # Confirm the mutually exclusive group has not been created
                if mutually_exclusive_name in pipeline_arg_groups:
                    raise Exception(
                        f"Cannot create mutually exclusive group {mutually_exclusive_name}. Group already exists."
                    )

                # Check if the group is required, if not given
                if "required" not in mutually_exclusive_args:
                    mutually_exclusive_args["required"] = False

                # Reassign the pipeline_arg_group based on required status
                if (
                    pipeline_arg_group == "basic"
                    and "required" in mutually_exclusive_args
                ):
                    mutually_exclusive_group = "required"
                elif pipeline_arg_group == "basic":
                    mutually_exclusive_group = "optional"
                else:
                    mutually_exclusive_group = pipeline_arg_group

                # Create the mutually exclusive group
                pipeline_arg_groups[mutually_exclusive_name] = pipeline_arg_groups[
                    mutually_exclusive_group
                ].add_mutually_exclusive_group(
                    required=mutually_exclusive_args["required"]
                )

            # Create dict to store the arg wildcard
            arg_wildcards = {}

            # Check if wildcards are present
            for wildcard_name, wildcards_args in self._yeildGroup(
                "wildcards-args", group_args
            ):
                # Process the default arg of the wildcard
                wildcards_args = self._processArgDefaults(wildcards_args)

                # Assign the argument wildcards, if applicable
                arg_wildcards[wildcard_name] = re.findall(
                    r"\{(.*?)\}", wildcards_args["default"]
                )

                # Add the argument to the args argument group
                group_args["args"][wildcard_name] = wildcards_args

            # Check if wildcards are present
            for pipeline_arg_name, arg_args in self._yeildGroup("args", group_args):
                # Process and arg type, if applicable
                if "type" in arg_args:
                    arg_args = self._processArgType(arg_args)

                # Process the arg action, if applicable
                if "action" in arg_args:
                    arg_args = self._processArgAction(arg_args)

                # Process the arg default, if applicable
                if "default" in arg_args:
                    arg_args = self._processArgDefaults(arg_args)

                # Assign the argument wildcards, if applicable
                if "wildcards" in arg_args:
                    arg_args = self._processArgWildcard(
                        pipeline_arg_name, arg_args, arg_wildcards
                    )

                # Check if a default value exists
                elif "default" in arg_args and not isinstance(
                    arg_args["default"], dict
                ):
                    # Add the default string to the help message
                    arg_args["help"] += f' (default: { arg_args["default"]})'

                # Reassign the pipeline_arg_group based on required status
                if "mutually-exclusive" in arg_args:
                    arg_group = arg_args["mutually-exclusive"]
                    del arg_args["mutually-exclusive"]
                elif pipeline_arg_group == "basic" and "required" in arg_args:
                    arg_group = "required"
                elif pipeline_arg_group == "basic":
                    arg_group = "optional"
                else:
                    arg_group = pipeline_arg_group

                # Assign the argument
                pipeline_arg_groups[arg_group].add_argument(
                    f"--{pipeline_arg_name}", **arg_args
                )

        # Add the common optional arguments, but at the end of the list
        pipeline_arg_groups["optional"].add_argument(
            "--workflow-prefix",
            help="Assign the workflow prefix",
            type=str,
            default=f"Workflow_{jobTimeStamp()}_{jobRandomString()}",
        )
        pipeline_arg_groups["optional"].add_argument(
            "--work-dir",
            help="Assign the working directory for snakemake. If not provided, the current directory will be used.",
            type=str,
        )
        pipeline_arg_groups["optional"].add_argument(
            "--scale-threads",
            help="Scale the threads for each task",
            type=float,
            default=1.0,
        )
        pipeline_arg_groups["optional"].add_argument(
            "--scale-mem",
            help="Scale the memory (RAM) for each task",
            type=float,
            default=1.0,
        )
        pipeline_arg_groups["optional"].add_argument(
            "--resource-yml",
            help="Create a seperate resource yaml",
            action="store_true",
        )
        pipeline_arg_groups["optional"].add_argument(
            "--singularity-dir",
            help="Assign different directory of singularity images",
            type=str,
        )
        pipeline_arg_groups["optional"].add_argument(
            "--no-overwrite",
            help="Do not overwrite existing files",
            action="store_false",
            dest="overwrite",
        )
        pipeline_arg_groups["optional"].add_argument(
            "-h", "--help", action="help", help="show this help message and exit"
        )

    def returnArgs(self):
        return vars(self._pipeline_parser.parse_args())

    @classmethod
    def create(cls, config_pipelines):
        return cls(config_pipelines)

    @staticmethod
    def _yeildGroup(group_name, group_args):
        if group_name in group_args:
            for group_name, group_dict in group_args[group_name].items():
                yield group_name, group_dict

    @staticmethod
    def _processArgType(arg_args):
        try:
            arg_args["type"] = locate(arg_args["type"])
        except:
            logging.exception(f"Unable to locate type: {arg_args['type']}")
            raise

        return arg_args

    @staticmethod
    def _processArgAction(arg_args):
        # Assign the confirmDir action
        if arg_args["action"] == "confirmDir":
            arg_args["action"] = confirmDir()

        # Assign the confirmFile action
        elif arg_args["action"] == "confirmFile":
            arg_args["action"] = confirmFile()

        return arg_args

    @staticmethod
    def _processArgDefaults(arg_args):
        # Skip if the default is not a dictionary
        if not isinstance(arg_args["default"], dict):
            return arg_args

        # Assign the default string
        default_str = arg_args["default"]["str"]

        # Process suffix strings
        if "suffix" in arg_args["default"]:
            # Check if the suffix is a string, then convert to a list
            if isinstance(arg_args["default"]["suffix"], str):
                arg_args["default"]["suffix"] = [arg_args["default"]["suffix"]]

            # Loop the suffixes
            for suffix in arg_args["default"]["suffix"]:
                # Add underscores if needed
                if default_str:
                    default_str += "_"

                # Process suffix strings
                if isinstance(suffix, str):
                    default_str += suffix

                # Process suffix dicts
                elif isinstance(suffix, dict):
                    # Convert the suffix dict to lowercase keys
                    suffix_dict = {_k.lower(): _v for _k, _v in suffix.items()}

                    if "function" not in suffix_dict:
                        raise Exception(f"Suffix dict not supported: {suffix_dict}")

                    if suffix["function"] == "jobTimeStamp":
                        default_str += jobTimeStamp()
                    elif suffix["function"] == "jobRandomString":
                        default_str += jobRandomString()
                    else:
                        raise Exception(f"Function not supported: {suffix['function']}")

        # Assign the default string
        arg_args["default"] = default_str

        return arg_args

    @staticmethod
    def _processArgWildcard(pipeline_arg_name, arg_args, arg_wildcards):
        # Check if the wildcards is valid
        if arg_args["wildcards"] not in arg_wildcards:
            raise Exception(
                f'Wildcards not found ({arg_args["wildcards"]}) for argument {pipeline_arg_name}'
            )

        # Confirm no default value is present
        if "default" in arg_args:
            raise Exception(
                f'Wildcards are incompatible with default values ({arg_args["default"]}) for argument {pipeline_arg_name}'
            )

        # Add the default string to the help message
        arg_args["help"] += (
            f' (supported wildcards: {", ".join(arg_wildcards[arg_args["wildcards"]])})'
        )

        del arg_args["wildcards"]

        return arg_args


# Create variables to store randomString and timeStamp
random_string = None
time_stamp = None



import pytest
import os
import tempfile

from pipemake.pathIO import DirIO


@pytest.mark.parametrize(
    "path_name",
    ["tests/files/pathIO/test_dir"],
)
def test_DirIO_exists_wo_error(path_name):
    DirIO.create(path_name)


@pytest.mark.parametrize(
    "path_name",
    ["tests/files/pathIO/test_dir2"],
)
def test_DirIO_exists_w_error(path_name):
    with pytest.raises(IOError):
        DirIO.create(path_name)


@pytest.mark.parametrize(
    "path_name",
    ["tests/files/pathIO/test_dir"],
)
def test_DirIO_standardize_wo_error(path_name):
    test_dir = tempfile.mkdtemp()
    test_Dir = DirIO.create(path_name)
    test_Dir.standardize("new_dir", workflow_prefix=test_dir)
    assert os.path.isdir(os.path.join(test_dir, "new_dir"))
    assert os.path.isfile(os.path.join(test_dir, "new_dir", "fasta_table.tsv"))

    with open(os.path.join(test_dir, "new_dir", "fasta_table.tsv")) as table_file:
        assert table_file.readline() == "genomes\tfilename\n"
        assert (
            table_file.readline() == "genome1\ttests/files/pathIO/test_dir/genome1.fa\n"
        )
        assert (
            table_file.readline() == "genome2\ttests/files/pathIO/test_dir/genome2.fa"
        )


@pytest.mark.parametrize(
    "path_name",
    ["tests/files/pathIO/test_dir"],
)
def test_DirIO_standardize_copy_wo_error(path_name):
    test_dir = tempfile.mkdtemp()
    test_Dir = DirIO.create(path_name)
    test_Dir.standardize("new_dir", workflow_prefix=test_dir, copy_method="copy")
    assert os.path.isdir(os.path.join(test_dir, "new_dir"))
    assert os.path.isfile(os.path.join(test_dir, "new_dir", "fasta_table.tsv"))

    with open(os.path.join(test_dir, "new_dir", "fasta_table.tsv")) as table_file:
        assert table_file.readline() == "genomes\tfilename\n"
        assert (
            table_file.readline() == "genome1\ttests/files/pathIO/test_dir/genome1.fa\n"
        )
        assert (
            table_file.readline() == "genome2\ttests/files/pathIO/test_dir/genome2.fa"
        )


@pytest.mark.parametrize(
    "path_name",
    ["tests/files/pathIO/test_dir"],
)
def test_DirIO_returnPaths_link_wo_error(path_name):
    test_Dir = DirIO.create(path_name)
    assert test_Dir.returnPaths() == [os.path.abspath(path_name)]


@pytest.mark.parametrize(
    "path_name",
    ["tests/files/pathIO/test_dir"],
)
def test_DirIO_returnPaths_copy_wo_error(path_name):
    test_Dir = DirIO.create(path_name)
    assert test_Dir.returnPaths(copy_method="copy") == []



import os
import re
import yaml
import shutil
import logging

from pathlib import Path
from collections import defaultdict

from pipemake.singularityIO import Singularity


class SnakePipelineIO:
    def __init__(
        self,
        workflow_prefix="",
        pipeline_storage_dir="",
        pipeline_job_dir="",
        work_dir="",
        singularity_dir="",
        resource_yml=None,
        scale_threads=0,
        scale_mem=0,
        indent_style="\t",
        overwrite=True,
        **kwargs,
    ):
        # Assign the basic arguments
        self._workflow_prefix = workflow_prefix

        # Assign the snakemake pipeline filename
        if workflow_prefix.endswith(".smk"):
            self.smkp_filename = workflow_prefix
        else:
            self.smkp_filename = f"{workflow_prefix}.smk"

        # Check if the overwrite is a bool or None
        if not isinstance(overwrite, bool):
            raise Exception(f"Invalid overwrite type: {overwrite}")

        # Confirm the file exists
        if os.path.isfile(self.smkp_filename) and not overwrite:
            raise IOError(
                f"SnakePipeline file already exists: {self.smkp_filename}. Please rename the SnakePipeline file or overwrite"
            )

        # Confirm the resource yml is bool or None
        if not isinstance(resource_yml, (bool, type(None))):
            raise Exception(f"Invalid resource yml type: {resource_yml}")

        # Confirm the scale threads and mem are integers
        if not isinstance(scale_threads, float):
            raise Exception(f"Invalid scale threads type: {scale_threads}")
        if not isinstance(scale_mem, float):
            raise Exception(f"Invalid scale mem type: {scale_mem}")

        # Confirm the indent style
        if indent_style not in [" ", "\t"]:
            raise Exception(f"Specified indent style not supported: {indent_style}")

        # Assign the basic arguments
        self._pipe_file = open(self.smkp_filename, "w")
        self._resource_yml = resource_yml
        self._scale_threads = scale_threads
        self._scale_mem = scale_mem
        self._mem_types = [
            "mem_b",
            "mem_kb",
            "mem_mb",
            "mem_gb",
            "mem_tb",
            "mem_pb",
            "mem_kib",
            "mem_mib",
            "mem_gib",
            "mem_tib",
            "mem_pib",
        ]
        self._indent_style = indent_style

        # Assign the module storage directory, and confirm it exists
        self._module_storage_dir = os.path.join(pipeline_storage_dir, "modules")
        if not os.path.exists(self._module_storage_dir):
            raise IOError(
                f"Unable to open: {self._module_storage_dir}. Please confirm the directory exists."
            )

        # Assign the script storage directory, and confirm it exists
        self._script_storage_dir = os.path.join(pipeline_storage_dir, "scripts")
        if not os.path.exists(self._script_storage_dir):
            raise IOError(
                f"Unable to open: {self._script_storage_dir}. Please confirm the directory exists."
            )

        # Assign the module config directory
        self._module_job_dir = os.path.join(pipeline_job_dir, "modules")
        if not os.path.exists(self._module_job_dir):
            os.makedirs(self._module_job_dir)

        # Assign the script config directory
        self._script_job_dir = os.path.join(pipeline_job_dir, "scripts")

        # Assign the backup directory
        self._backup_dir = os.path.join(pipeline_job_dir, "backups")
        if not os.path.exists(self._backup_dir):
            os.makedirs(self._backup_dir)

        # Assign the work directory
        self._work_dir = work_dir

        # Create arguments to populate
        self._module_filenames = []
        self._output_list = []
        self._config_params = set()
        self._resource_params = defaultdict(lambda: defaultdict(int))
        self._singularity_dir = singularity_dir
        self._pipeline_singularity_dict = {}

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @classmethod
    def open(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def addModule(self, module_filename, **kwargs):
        # Update the module filename, if necessary
        if not module_filename.endswith(".smk"):
            module_filename = f"{module_filename}.smk"

        # Create the filepath to the stored module
        storage_module_path = os.path.join(self._module_storage_dir, module_filename)

        # Confirm the stored module exists
        if not os.path.isfile(storage_module_path):
            raise IOError(
                f"Unable to open: {storage_module_path}. Please confirm the file exists."
            )

        # Open and process the stored module file
        smk_module = SnakeFileIO.open(
            storage_module_path, singularity_dir=self._singularity_dir
        )

        # Check if the module has script files
        if len(smk_module._file_script_files) > 0:
            # Create the script job directory
            if not os.path.exists(self._script_job_dir):
                os.makedirs(self._script_job_dir)

        # Loop the rules to assign the output
        for rule in smk_module._file_rules:
            # Skip the rule if not the output rule
            if rule.rule_name != "all":
                continue

            # Check if the rule has output
            if rule._rule_output_list:
                # Save the output list from the rule
                for rule_output in rule._rule_output_list:
                    self._output_list.append(
                        _convert_indent(
                            rule_output, rule._indent_style, self._indent_style
                        )
                    )

        # Loops the file rules in test
        for rule in smk_module._file_rules:
            # Save the config params from the rule
            self._config_params.update(rule._rule_config_params)

            # Scale and save the resource params
            for resource_type, resource_value in rule._rule_resource_params.items():
                if resource_type == "threads" and self._scale_threads:
                    resource_value = self._scaler(resource_value, self._scale_threads)
                elif (
                    "mem_" in resource_type
                    and resource_type.lower() in self._mem_types
                    and self._scale_mem
                ):
                    resource_value = self._scaler(resource_value, self._scale_mem)
                self._resource_params[rule.rule_name][resource_type] = resource_value

        # Create the job module filepath
        storage_module_path = os.path.join(self._module_job_dir, module_filename)

        # Create the job module file
        smk_module.write(storage_module_path)

        # Add the module filename to the list
        self._module_filenames.append(module_filename)

        # Store the singularity containers from the module
        for (
            singularity_path,
            singularity_container,
        ) in smk_module._file_singularity_dict.items():
            if singularity_path not in self._pipeline_singularity_dict:
                self._pipeline_singularity_dict[singularity_path] = (
                    singularity_container
                )

        for script_file in smk_module._file_script_files:
            shutil.copy(
                os.path.join(self._script_storage_dir, script_file),
                os.path.join(self._script_job_dir, script_file),
            )
            logging.info(f"Script added to pipeline: {script_file}")

        logging.info(f"Module added to pipeline: {module_filename}")

    def buildSingularityContainers(self):
        # Loop the singularity containers
        for singularity_container in self._pipeline_singularity_dict.values():
            # Download the singularity container
            singularity_container.download()

    def buildSingularityURLTable(self):
        # Create file in the backup directory
        url_table_file = open(
            os.path.join(self._backup_dir, "singularity_urls.tsv"), "w"
        )

        # Write the header
        url_table_file.write("Singularity Path\tURL\n")

        # Loop the singularity containers
        for singularity_container in self._pipeline_singularity_dict.values():
            # Write the singularity path and URL
            url_table_file.write(
                f"{singularity_container._image_filename}\t{singularity_container._url}\n"
            )

        # Close the file
        url_table_file.close()

    def writeConfig(self, pipeline_args):
        # Create yml dicts
        yml_config_dict = {}
        yml_resource_dict = {"resources": {}}

        # Check if the workdir is in the pipeline args
        if self._work_dir:
            yml_config_dict["workdir"] = self._work_dir

        # Create a list of config params to sort by length
        config_params_list = list(self._config_params)
        config_params_list.sort(key=lambda t: len(t))

        # Populate the config yml dict with the pipeline args
        for config_param in config_params_list:
            # Check if the config param is within a group
            if len(config_param) > 1:
                group, config_arg = config_param

            # Assign the config arg from the config param
            else:
                group = None
                config_arg = config_param[0]

            # Create the equivalent pipeline arg for the config arg
            pipeline_arg = config_arg.replace("-", "_")

            # Raise an error if the pipeline arg is not in the pipeline args
            if pipeline_arg not in pipeline_args:
                raise Exception(f"Pipeline arg not found: {pipeline_arg}")

            # Check if the group is not in the yml dict
            if group and group not in yml_config_dict:
                yml_config_dict[group] = {}

            if not group:
                yml_config_dict[config_arg] = pipeline_args[pipeline_arg]
            else:
                yml_config_dict[group][config_arg] = pipeline_args[pipeline_arg]

        # Populate the resource yml dict
        for rule_name, rule_resource_dict in self._resource_params.items():
            if not rule_resource_dict:
                continue

            # Create the rule resource dict
            yml_resource_dict["resources"][rule_name] = {}

            # Populate the rule resource dict
            for resource_name, resource_value in rule_resource_dict.items():
                yml_resource_dict["resources"][rule_name][resource_name] = (
                    resource_value
                )

        # Check if the resource yml should be a separate file
        if not self._resource_yml:
            yml_config_dict.update(yml_resource_dict)
        else:
            self._createYml(yml_resource_dict, f"{self._workflow_prefix}.resources.yml")

        self._createYml(yml_config_dict, f"{self._workflow_prefix}.yml")

    def writePipeline(self):
        # Assign the config file(s) and working directory
        self._pipe_file.write(f"configfile: '{self._workflow_prefix}.yml'\n")
        if self._resource_yml:
            self._pipe_file.write(
                f"configfile: '{self._workflow_prefix}.resources.yml'\n"
            )
        if self._work_dir:
            self._pipe_file.write("\nworkdir: config['workdir']\n\n")

        # Create the rule all block
        self._pipe_file.write("rule all:\n")
        self._pipe_file.write(f"{self._indent_style}input:\n")

        # Add the rule all input
        indented_input = "\n".join(_output for _output in self._output_list)
        self._pipe_file.write(indented_input + "\n\n")

        # Create the input block of modules
        for module_filename in self._module_filenames:
            self._pipe_file.write(
                f'include: "{os.path.join(self._module_job_dir, module_filename)}"\n'
            )

        logging.info("Pipeline written successfully")

    def close(self):
        self._pipe_file.close()

        # Create the singularity url table, if necessary
        if self._pipeline_singularity_dict:
            self.buildSingularityURLTable()

        # Assign the basename of the workflow prefix
        workflow_basename = os.path.basename(self._workflow_prefix)

        # Create backups of the pipeline files
        shutil.copy(
            f"{self._workflow_prefix}.smk",
            os.path.join(self._backup_dir, f"{workflow_basename}.smk.bkp"),
        )
        shutil.copy(
            f"{self._workflow_prefix}.yml",
            os.path.join(self._backup_dir, f"{workflow_basename}.yml.bkp"),
        )
        if os.path.isfile(f"{self._workflow_prefix}.resources.yml"):
            shutil.copy(
                f"{self._workflow_prefix}.resources.yml",
                os.path.join(
                    self._backup_dir, f"{workflow_basename}.resources.yml.bkp"
                ),
            )

        logging.info("Pipeline backups created successfully")

    @staticmethod
    def _scaler(value, scaler, value_type=float, output_type=int, **kwargs):
        # Confirm a scaler was given
        if not scaler:
            raise Exception(f"No scaler given for: {value}")

        # Convert the value to the correct type
        try:
            value = value_type(value)
        except:
            logging.exception(f"Unable to convert value to type: {value}")
            raise

        # Scale and return the input
        return output_type(value * scaler)

    @staticmethod
    def _createYml(yml_dict, yml_filename, **kwargs):
        class MyDumper(yaml.Dumper):
            def increase_indent(self, flow=False, indentless=False):
                return super(MyDumper, self).increase_indent(flow, False)

        yml_file = open(yml_filename, "w")
        yml_file.write(yaml.dump(yml_dict, Dumper=MyDumper, sort_keys=False))
        yml_file.close()


class SnakeFileIO:
    def __init__(self, smk_filename, singularity_dir="", **kwargs):
        # Confirm the file exists
        if not os.path.isfile(smk_filename):
            raise IOError(
                f"Unable to open: {smk_filename}. Please confirm the file exists."
            )

        # Assign the basic arguments
        self.filename = smk_filename
        self._singularity_dir = singularity_dir
        self._indent_style = None
        self._output_rule = "all"
        self._exclude_rules = [self._output_rule]
        self._file_rules = []
        self._non_rule_text = ""

        # Assign the indent style
        self._assignIndent()

        # Parse the snakefile
        self._parseFile()

    @property
    def _file_singularity_dict(self):
        file_singularity_dict = {}
        for rule in self._file_rules:
            for (
                singularity_path,
                singularity_container,
            ) in rule._rule_singularity_dict.items():
                if singularity_path not in file_singularity_dict:
                    file_singularity_dict[singularity_path] = singularity_container
        return file_singularity_dict

    @property
    def _file_script_files(self):
        file_script_files = []
        for rule in self._file_rules:
            for script_file in rule._rule_script_files:
                file_script_files.append(script_file)
        return file_script_files

    def _assignIndent(self):
        # Bool to store if within a rule
        in_rule_block = False

        with open(self.filename) as smk_file:
            for smk_line in smk_file:
                # Skip blank lines
                if not smk_line.strip():
                    continue

                # Confirm indent style once assigned
                if self._indent_style:
                    # Skip rules as they have no indent
                    if (
                        smk_line.startswith("rule")
                        or smk_line.startswith("checkpoint")
                        or smk_line.startswith("def")
                    ):
                        continue

                    if not smk_line.startswith(self._indent_style):
                        raise Exception(
                            f'Inconsistent indent style "{smk_line[0]}" in: {self.filename}'
                        )

                    # Get the count of the indent style
                    indent_count = len(smk_line) - len(smk_line.lstrip())

                    # Calc the indent remainder
                    indent_remainder = indent_count % len(self._indent_style)

                    # Check indent remainder
                    if indent_remainder:
                        raise Exception(f"Indent style remainder: {indent_remainder}")

                # Assign the indent style
                elif in_rule_block:
                    # Get the count of the leading whiespace
                    leading_whitespace_count = len(smk_line) - len(smk_line.lstrip())

                    # Check for table indent style
                    if smk_line[0] == "\t":
                        self._indent_style = "\t" * leading_whitespace_count
                        logging.info(
                            f"Indent style: Tab(s). Count: {leading_whitespace_count}"
                        )

                    # Check for space indent style
                    elif smk_line[0] == " ":
                        self._indent_style = " " * leading_whitespace_count
                        logging.info(
                            f"Indent style: Spaces(s). Count: {leading_whitespace_count}"
                        )

                    # Report error, if other
                    else:
                        raise Exception("Unable to assign indent style")

                # Check if within first rule block
                if smk_line.split(" ")[0] == "rule":
                    in_rule_block = True

    def _parseFile(self):
        # Create a string to store the rule block
        rule_block = ""

        # Create bool to check if in the config block
        within_rule_block = False

        # Parse the snakefile
        with open(self.filename) as smk_file:
            for smk_line in smk_file:
                # Skip blank lines
                if not smk_line.strip():
                    continue

                # Split the line by the indent style (rule, attribute, param)
                smk_list = smk_line.rstrip().split(self._indent_style)

                # Assign if line is a rule or checkpoint
                is_rule_line = (
                    smk_list[0].startswith("rule")
                    or smk_list[0].startswith("checkpoint")
                ) and smk_line.rstrip().endswith(":")

                # Check if the line is the end of a rule
                if smk_list[0] and not is_rule_line:
                    within_rule_block = False

                # Check if the line is the start of a rule
                if is_rule_line:
                    within_rule_block = True

                    # Check if a previous rule block was found
                    if rule_block:
                        self._file_rules.append(
                            SnakeRuleIO.read(
                                rule_block,
                                singularity_dir=self._singularity_dir,
                                indent_style=self._indent_style,
                                output_rule=self._output_rule,
                            )
                        )
                        rule_block = ""

                # Add the line to the rule block
                if within_rule_block and smk_line.strip():
                    rule_block += smk_line

                # Save text that is not within a rule
                if not within_rule_block and smk_line.strip():
                    self._non_rule_text += smk_line

            # Parse the final rule block
            if rule_block:
                self._file_rules.append(
                    SnakeRuleIO.read(
                        rule_block,
                        singularity_dir=self._singularity_dir,
                        indent_style=self._indent_style,
                        output_rule=self._output_rule,
                    )
                )

    def write(self, filename):
        # Open the file
        with open(filename, "w") as smk_file:
            # Check if there is any non-rule text
            if self._non_rule_text:
                smk_file.write(self._non_rule_text)

            # Loop the file rules
            for rule in self._file_rules:
                # Write the rule, if included in the output
                if rule._output_rule:
                    smk_file.write(f"\n\n{rule}")

    @classmethod
    def open(cls, *args, **kwargs):
        return cls(*args, **kwargs)


class SnakeRuleIO:
    def __init__(
        self, rule_str, singularity_dir="", indent_style=None, output_rule="", **kwargs
    ):
        # Assign the rule argument to populate
        self._original_text = rule_str
        self._original_text_list = self._original_text.splitlines()
        self.rule_name = self._original_text_list[0].split()[1][:-1]
        self._output_rule = False if self.rule_name == output_rule else True
        self._singularity_dir = singularity_dir
        self._indent_style = indent_style
        self._rule_config_params = set()
        self._rule_singularity_dict = {}
        self._rule_script_files = []
        self._rule_resource_params = {}
        self._rule_text = self._original_text_list[0] + "\n"
        self._rule_output_list = []

        # Assign the fixed arguments
        self._resource_attributes = ["threads", "resources"]

        # Update and assign the rule parameters
        if self._output_rule:
            self._updateRule(self._original_text_list[1:])
            self._setParams()

        # Set the output list
        elif not self._output_rule:
            self._setOutput(self._original_text_list[1:])
            self._setParams()

    def __str__(self):
        return self._rule_text

    def __repr__(self):
        return self._rule_text

    @classmethod
    def read(cls, rule_str, *args, **kwargs):
        return cls(rule_str, *args, **kwargs)

    def _updateResource(self, resource_attribute, rule_line, attribute_level):
        # Assign the resource str
        resource_str = rule_line.strip()

        # Check if the resource is a plain integer
        if resource_str.isdigit():
            resource_name = resource_attribute
            resource_value = int(resource_str)

        else:
            # Assign the expected resource syntax
            resource_syntax_dict = {1: ":", 2: "="}

            # Check if the resource attributes can be assigned
            if resource_syntax_dict[attribute_level] not in rule_line:
                raise Exception(
                    f"Resource attribute error for {resource_attribute}. Value not assigned: {rule_line}"
                )

            # Assign the resource name and value
            resource_name, resource_value = rule_line.strip().split(
                resource_syntax_dict[attribute_level]
            )

            # Remove any whitespaces
            resource_name = resource_name.strip()
            resource_value = resource_value.strip()

            # Remove any trailing commas
            if resource_value.endswith(","):
                resource_value = resource_value[:-1]

            # Determine if the resource value type
            if not resource_value.isdigit():
                raise Exception(
                    f"Complex resource support not yet implemented: {rule_line}"
                )

        # Assign the resource value
        self._rule_resource_params[resource_name] = int(resource_value)

        # Assign the resource config attribute
        resource_config = f'config["resources"]["{self.rule_name}"]["{resource_name}"]'

        # Update the resource list
        rule_line = rule_line.replace(str(resource_value), resource_config)

        return rule_line

    def _updateContainer(self, container_str):
        # Check if a singularity directory is specified
        if not self._singularity_dir:
            return container_str

        # Assign the container text
        container_url = container_str.strip().split('"')[1]

        # Store the container
        container = Singularity.fromURL(container_url, self._singularity_dir)

        # Add the container to the singularity dict
        self._rule_singularity_dict[container.returnPath()] = container

        return container.updateContainer(container_str)

    def _updateScript(self, script_str, script_dir="scripts"):
        # Assign the script path
        script_path = Path(script_str.strip().split('"')[1])

        # Add the script file to the list
        self._rule_script_files.append(script_path.name)

        # Assign the script path
        script_output_path = os.path.join("..", script_dir, script_path.name)

        # Return the container path
        return script_str.split('"')[0] + f'"{script_output_path}"'

    def _setParams(self):
        # Find all occurences of the word config follpwed by one or more sets of square brackets
        for config_match in re.finditer(r"config(\[[^\]]*\])+", self._original_text):
            # Get the config string by removing newlines and whitespaces
            config_str = "".join(
                [_s.strip() for _s in config_match.group(0).splitlines()]
            )

            # Split the param str by either square brackets using regex
            config_list = [
                _p
                for _p in re.split(
                    r"\[|\]", config_str.replace('"', "").replace("'", "")
                )
                if _p
            ][1:]

            # Check if the match is associated with a value
            config_has_value = (
                False if self._original_text.find(f"{config_str}:") == -1 else True
            )

            # Check if the param does not have a value, i.e. config argument
            if config_has_value:
                raise Exception(f"Config param error. Has value: {self._original_text}")

            # Confirm the config list isn't more than two parameters
            if len(config_list) > 2:
                raise Exception(
                    f"Config param error. Too many param levels: {self._original_text}"
                )

            # Add the config assignment to the set
            self._rule_config_params.add(tuple(config_list))

    def _updateRule(self, rule_list):
        for rule_line in rule_list:
            if not rule_line.strip():
                continue

            # Split the line by the indent style
            rule_line_list = rule_line.rstrip().split(self._indent_style)

            # Assign the attribute level
            attribute_level = _attributeLevel(rule_line_list)

            # Check if the line is a rule attribute
            if attribute_level == 1:
                attribute_name, attribute_value = [
                    _a.strip() for _a in rule_line_list[1].split(":", 1)
                ]

            # Process resource attributes
            if attribute_name in self._resource_attributes:
                if attribute_value:
                    rule_line = self._updateResource(
                        attribute_name, rule_line, attribute_level
                    )
                elif attribute_level > 1:
                    rule_line = self._updateResource(
                        attribute_name, rule_line, attribute_level
                    )

            # Process container attributes
            elif attribute_name == "singularity":
                if attribute_value:
                    rule_line = self._updateContainer(rule_line)
                elif attribute_level > 1:
                    rule_line = self._updateContainer(rule_line)

            # Process script attributes
            elif attribute_name == "script":
                if attribute_value:
                    rule_line = self._updateScript(rule_line)
                elif attribute_level > 1:
                    rule_line = self._updateScript(rule_line)

            # Update the rule text
            self._rule_text += rule_line + "\n"

    def _setOutput(self, rule_list):
        for rule_line in rule_list:
            if not rule_line.strip():
                continue

            # Split the line by the indent style
            rule_line_list = rule_line.rstrip().split(self._indent_style)

            # Assign the attribute level
            attribute_level = _attributeLevel(rule_line_list)

            # Check if the line is a rule attribute
            if attribute_level == 1 and rule_line.rstrip().endswith(":"):
                attribute_name = rule_line_list[1][:-1]

            # Check if the rule is output, confirm the attribute is input, and update the rule output
            elif attribute_name == "input":
                self._rule_output_list.append(rule_line.rstrip())

        # Check that output list was populated
        if not self._rule_output_list:
            raise Exception("Output list not populated for rule")


def _attributeLevel(attribute_list):
    for attribute_level, attribute_item in enumerate(attribute_list):
        if not attribute_item:
            continue
        return attribute_level


def _convert_indent(source_str, source_indent_style, target_indent_style):
    # Get the count of the source indent style
    indent_count = source_str.count(source_indent_style)
    # Return the converted indent style
    return f"{indent_count*target_indent_style}{source_str.strip()}"



import os
import gzip
import shutil
import logging

import pandas as pd

from collections import defaultdict

from snakemake.io import get_wildcard_names


def checkIfGzipped(filename):
    with open(filename, "rb") as check_file:
        check_line = check_file.readline()
        bgzip_header = (
            b"\x1f\x8b\x08\x04\x00\x00\x00\x00\x00\xff\x06\x00\x42\x43\x02\x00"
        )
        gzip_header = b"\x1f\x8b"
        if check_line[: len(bgzip_header)] == bgzip_header:
            return True
        if check_line[: len(gzip_header)] == gzip_header:
            return True
    return False


class FileIO:
    def __init__(self, seq_filename, **kwargs):
        # Confirm the file exists
        if not os.path.isfile(seq_filename):
            raise IOError(
                f"Unable to open: {seq_filename}. Please confirm the file exists."
            )

        # Convert the filename to an absolute path
        abs_seq_filename = os.path.abspath(seq_filename)

        # Confirm the file using the absolute path
        if not os.path.isfile(abs_seq_filename):
            raise IOError(
                f"Unable to open: {abs_seq_filename}. Please confirm the file exists."
            )

        # Assign the basic arguments
        self.filename = abs_seq_filename

        # Assign if the file is compressed
        self.gzipped = checkIfGzipped(self.filename)

        # Stores path if using a link
        self.link_path = ""

    @classmethod
    def create(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    def standardize(
        self,
        standardized_filename,
        out_dir="",
        workflow_prefix="",
        work_dir="",
        gzipped=None,
        copy_method="symbolic_link",
        **kwargs,
    ):
        # Assign the destination filename
        dest_filename = standardized_filename

        # Create path as needed
        if out_dir:
            dest_filename = os.path.join(out_dir, dest_filename)
        if workflow_prefix:
            dest_filename = os.path.join(workflow_prefix, dest_filename)
        if work_dir:
            dest_filename = os.path.join(work_dir, dest_filename)

        # Create the output directory, if needed
        if not os.path.exists(os.path.dirname(dest_filename)):
            os.makedirs(os.path.dirname(dest_filename))

        # Copy files with the same gzipped status
        if gzipped is None or self.gzipped == gzipped:
            if copy_method == "symbolic_link":
                self.link_path = os.path.dirname(self.filename)
                os.symlink(self.filename, dest_filename)
            elif copy_method == "copy":
                shutil.copy(self.filename, dest_filename)
            else:
                raise Exception(f"Unknown copy method: {copy_method}")

        # Copy to a decompressed status
        elif self.gzipped and not gzipped:
            with gzip.open(self.filename, "r") as f_in, open(
                dest_filename, "wb"
            ) as f_out:
                shutil.copyfileobj(f_in, f_out)

        # Copy to a compressed status
        elif not self.gzipped and gzipped:
            with open(self.filename, "r") as f_in, gzip.open(
                dest_filename, "wt"
            ) as f_out:
                shutil.copyfileobj(f_in, f_out)

        # Raise exception (not possible?)
        else:
            raise Exception(f"Unable to copy file: {self.filename}")

        logging.info(
            f"Standardized: {self.filename} to {dest_filename}. Method: {copy_method}"
        )

    def returnPaths(self, copy_method="symbolic_link", **kwargs):
        if copy_method == "copy":
            return []
        elif copy_method == "symbolic_link":
            return [os.path.dirname(self.filename)]
        else:
            raise Exception(f"Unsupported copy method: {copy_method}")


class TableIO:
    def __init__(self, table_dataframe, sample_column="", **kwargs):
        # Confirm the table is a pandas DataFrame
        if not isinstance(table_dataframe, pd.DataFrame):
            raise Exception(
                f"Table must be stored as a pandas DataFrame: {table_dataframe}"
            )

        # Assign the basic arguments
        self._table_dataframe = table_dataframe
        self._sample_column = sample_column
        self._file_columns = set()
        self._table_columns = set([self._sample_column])
        self.samples = defaultdict(list)

        if self._sample_column not in self._table_dataframe.columns:
            raise Exception(
                f"Unable to assign sample column ({self._sample_column}) in DataFrame columns: {self._table_dataframe.columns}"
            )

        # Assign arguments from the columns of the dataframe
        for col in self._table_dataframe.columns:
            if col == self._sample_column:
                self.samples[col] = list(self._table_dataframe[self._sample_column])
            elif "filename" in col:
                self._file_columns.add(col)
            elif ":" in col:
                wildcard_item, sample_item = col.split(":")
                self.samples[wildcard_item].append(sample_item)
                self._file_columns.add(wildcard_item)
                self._table_columns.add(wildcard_item)
            else:
                self._table_columns.add(col)

        # Check the samples are unique
        for sample_list in self.samples.values():
            if len(sample_list) != len(list(set(sample_list))):
                raise Exception("Sample names are not unique")

        if len(self._file_columns) > 1:
            raise Exception("Files may exist within a single column")

        # Create a set of all the columns
        # self._table_columns = self._table_columns.union(self._file_columns)

    @property
    def _file_column(self):
        if len(self._file_columns) > 1:
            raise Exception("Files may exist within a single column")
        return list(self._file_columns)[0]

    @classmethod
    def fromFilenameStr(cls, table_filename, sep="\t", **kwargs):
        return cls(
            table_dataframe=pd.read_csv(
                table_filename, sep=sep, index_col=False, dtype=str
            ).fillna(""),
            **kwargs,
        )

    def standardizedFiles(self, standardized_wildcard, **kwargs):
        # Skip if nn file columns
        if not list(self._file_columns):
            return

        # Assign the standardized wildcard names
        standardized_wildcard_names = get_wildcard_names(standardized_wildcard)

        # Confirm the dataframe column are being standardized
        for col in self._table_columns:
            if col not in standardized_wildcard_names:
                raise Exception(
                    f"Unable to assign column ({col}) for standardization: {standardized_wildcard}"
                )
            standardized_wildcard_names.remove(col)

        # Confirm all wildcards were used
        if len(standardized_wildcard_names) > 0:
            raise Exception(
                f"Not all wildcards found in table: {', '.join(list(standardized_wildcard_names))}"
            )

        # Loop the dataframe by row
        for _, row_dict in self._table_dataframe.iterrows():
            # Assign the non-filename wildcards for the row
            str_wildcards = {
                _c: _v for _c, _v in row_dict.items() if self._file_column not in _c
            }

            # Loop the row columns (i.e. indices)
            for file_col, sample_filename in row_dict.items():
                # Confirm the row column (i.e. index) is a filename
                if self._file_column not in file_col:
                    continue

                # Skip if sample filename is blank (e.g. No R2 filename)
                if not sample_filename:
                    continue

                # Confirm the file exists
                if not os.path.isfile(sample_filename):
                    raise IOError(
                        f"Unable to locate file. Column: {file_col}. Filename: {sample_filename}"
                    )

                # Check if the file column is a wildcard
                if ":" in file_col:
                    # Create the wilcard dict for the sample
                    file_wildcard_dict = dict([file_col.split(":")])
                    str_wildcards.update(file_wildcard_dict)
                    # file_wildcard_dict.update(str_wildcards)

                # Create the standardized filename
                standardized_filename = standardized_wildcard.format(**str_wildcards)

                # Standardize the file
                sample_file = FileIO.create(sample_filename)
                sample_file.standardize(standardized_filename, **kwargs)

    def returnPaths(self, copy_method="symbolic_link", **kwargs):
        if copy_method == "copy":
            return []
        elif copy_method == "symbolic_link":
            # Create list to store paths
            path_set = set()

            # Loop the dataframe by row
            for _, row_dict in self._table_dataframe.iterrows():
                for file_col, sample_filename in row_dict.items():
                    # Confirm the row column (i.e. index) is a filename
                    if self._file_column not in file_col:
                        continue

                    # Skip if sample filename is blank (e.g. No R2 filename)
                    if not sample_filename:
                        continue

                    # Confirm the file exists
                    if not os.path.isfile(sample_filename):
                        raise IOError(
                            f"Unable to locate file. Column: {file_col}. Filename: {sample_filename}"
                        )

                    # Assign and store the directory
                    path_name = os.path.dirname(sample_filename)
                    if not os.path.isdir(path_name):
                        raise Exception(f"Unable to assign path: {path_name}")
                    path_set.add(os.path.abspath(path_name))

            return list(path_set)

        else:
            raise Exception(f"Unsupported copy method: {copy_method}")

    def returnSamples(self):
        return self.samples



import io
import os

from setuptools import setup

import pipemake

with io.open("README.rst", "rt", encoding="utf8") as f:
    readme = f.read()

# Read requirements.txt
lib_folder = os.path.dirname(os.path.realpath(__file__))
requirement_path = lib_folder + "/requirements.txt"
requirements = []
if os.path.isfile(requirement_path):
    with open(requirement_path) as f:
        requirements = f.read().splitlines()

setup(
    name=pipemake.__name__,
    version=pipemake.__version__,
    project_urls={
        "Documentation": pipemake.__docs__,
        "Code": pipemake.__code__,
        "Issue tracker": pipemake.__issue__,
    },
    license=pipemake.__license__,
    url=pipemake.__url__,
    description=pipemake.__summary__,
    long_description_content_type="text/x-rst",
    long_description=readme,
    packages=["pipemake"],
    install_requires=requirements,
    entry_points={
        "console_scripts": [
            "pipemake=pipemake.pipemake:main",
        ],
    },
    python_requires=">=3.7",
)



import os
import logging
import itertools

from collections import defaultdict
from snakemake.io import glob_wildcards

from pipemake.fileIO import FileIO


class WildcardIO:
    def __init__(self, wildcard_str, wildcard_dict, sample_wildcards=[], **kwargs):
        # Confirm wildcards were assigned
        if not list(wildcard_dict):
            raise Exception(f"Unable to find wildcards within: {wildcard_str}")

        # Assign the basic arguments
        self.wildcard_str = wildcard_str
        self.wildcard_dict = wildcard_dict

        # Check if the wildcard dict is empty
        missing_wildcards = [
            _w_name for _w_name, _w_value in self.wildcard_dict.items() if not _w_value
        ]
        if missing_wildcards:
            # Report that all wildcards are missing, likely due to incorrect filenames
            if len(missing_wildcards) == len(self.wildcard_dict):
                raise Exception(
                    "Unable to assign by wildcards. Please confirm the filenames are correct."
                )

            # Report the missing wildcards, not sure if this is possible
            else:
                raise Exception(
                    f"Unable to find wildcard values for: {', '.join(missing_wildcards)}"
                )

        if not sample_wildcards:
            self.samples = {}
        else:
            self.samples = {_col: wildcard_dict[_col] for _col in sample_wildcards}

    @classmethod
    def fromStr(cls, wildcard_str, **kwargs):
        # Create a dictionary to store the wildcard values
        wildcard_dict = defaultdict(list)

        # Use glob to retrieve wildcard entries
        wildcard_data = glob_wildcards(wildcard_str)

        # Loop the wildcard names - i.e. fields
        for wildcard_name in wildcard_data._fields:
            # Assign the wildcard values
            wildcard_values = list(dict.fromkeys(getattr(wildcard_data, wildcard_name)))

            # Loop the wildcard values
            for wildcard_value in wildcard_values:
                # Skip if the wildcard value is not a path
                if os.sep in wildcard_value:
                    # Assign the possible wildcard directory
                    wildcard_directory = os.path.join(
                        os.path.dirname(wildcard_str), os.path.dirname(wildcard_value)
                    )

                    # Check if the wildcard value contains a directory
                    if os.path.isdir(wildcard_directory):
                        logging.warning(
                            f"Wildcard ({wildcard_name}) value contains a directory: {os.path.dirname(wildcard_value)}. This is often caused by a subdirectory in the wildcard directory."
                        )
                        continue

                # Append the wildcard value to the wildcard dict
                wildcard_dict[wildcard_name].append(wildcard_value)

        return cls(wildcard_str, wildcard_dict, **kwargs)

    def standardizedFiles(self, standardized_wildcard, **kwargs):
        # Create list of the wildcard name and values
        wildcard_names, wildcard_values = zip(*self.wildcard_dict.items())

        # Format the wildcard str for the sample and standardized file
        for sample_wildcard_dict in [
            dict(zip(wildcard_names, v)) for v in itertools.product(*wildcard_values)
        ]:
            sample_filename = self.wildcard_str.format(**sample_wildcard_dict)

            # Confirm the sample file exists
            if not os.path.isfile(sample_filename):
                logging.warning(
                    f"Unable to find file: {sample_filename}. This may occur if wildcards are not always combined."
                )
                continue

            # Check if any standardized wildcards are missing
            missing_wildcards = [
                _w
                for _w in glob_wildcards(standardized_wildcard)._fields
                if _w not in sample_wildcard_dict
            ]

            # Confirm the wildcards provided are in the sample wildcard dict
            if missing_wildcards:
                raise Exception(
                    f"Unable to assign [{', '.join(missing_wildcards)}] in {standardized_wildcard}. Please confirm the same wildcards are used in the sample and standardized filenames."
                )

            standardized_filename = standardized_wildcard.format(**sample_wildcard_dict)

            # Standardize the file
            sample_file = FileIO.create(sample_filename)
            sample_file.standardize(standardized_filename, **kwargs)

    def returnPaths(self, copy_method="symbolic_link", **kwargs):
        if copy_method == "copy":
            return []
        elif copy_method == "symbolic_link":
            path_name = os.path.dirname(self.wildcard_str)
            if not os.path.isdir(path_name):
                raise Exception(f"Unable to assign path: {path_name}")
            return [os.path.abspath(path_name)]
        else:
            raise Exception(f"Unsupported copy method: {copy_method}")

    def returnSamples(self):
        return self.samples
